# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module hosts OS-specific code common to all Unix-like OSes."""

from image_creator.os_type import OSBase, sysprep, add_sysprep_param

# Credits go to Wmconfig (https://www.arrishq.net/) for the biggest part of
# this collection
DESKTOPSESSIONS = {
    # Put the most popular GUIs first
    ('startkde',): 'KDE',
    ('gnome-session', 'gnomesession'): 'GNOME',
    ('unity',): 'Unity',
    ('xfce4-session', 'startxfce4', 'xfwm4'): 'Xfce',
    ('startlxde', 'lxsession'): 'LXDE',
    ('mate-session',): 'MATE',
    ('cinnamon', 'gnome-session-cinnamon', 'cinnamon-session'): 'Cinnamon',
    ('enlightenment',): 'Enlightenment',

    ('aewm',): 'aewm',
    ('afterstep',): 'AfterStep',
    ('amiwm',): 'amiwm',
    ('awesome',): 'awesome',
    ('blackbox',): 'Blackbox',
    ('ctwm',): 'CTWM',
    ('dwm',): 'dwm',
    ('epiwm',): 'EPIwm',
    ('fluxbox', 'startfluxbox'): 'Fluxbox',
    ('flwm',): 'flvm',
    ('fvwm',): 'FVWM',
    ('fvwm2',): 'FVWM',
    ('fvwm95',): 'FVWM95',
    ('golem',): 'Golem',
    ('i3',): 'i3',
    ('icewm', 'icewm-session'): 'IceVM',
    ('ion', 'ion2', 'ion3'): 'Ion',
    ('jwm',): "JWM",
    ('kahakai',): 'Kahakai',
    ('larswm',): 'larswm',
    ('mlvwm',): 'MLVWM',
    ('mwm',): 'Motif Window Manager',
    ('olvwm',): 'OLVWM',
    ('olwm',): 'OLWM',
    ('openbox', 'openbox-session'): 'Openbox',
    ('particleman', ): 'ParticleMan',
    ('pekwm',): 'PeKWM',
    ('pwm', 'pwm2', 'pwm3'): 'PWM',
    ('qvwm',): 'qvwm',
    ('rasor-session',): 'Razor-qt',
    ('ratpoison',): 'Ratpoison',
    ('sapphire',): 'Sapphire',
    ('sawfish',): 'Sawfish',
    ('sithwm',): 'SithWM',
    ('startede',): 'EDE',
    ('startstump',): 'StumpWM',
    ('starttrinity',): 'TDE',
    ('twm',): 'twm',
    ('uwm',): 'UDE',
    ('w9wm',): 'w9wm',
    ('wmaker',): 'WindowMaker',
    ('wmx',): 'wmx',
    ('wmi', 'wmii'): 'wmii',
    ('windowlab',): 'WindowLab',
    ('xmonad',): 'xmonad',
}
X11_EXECUTABLE = 'startx'

SENSITIVE_USERDATA = [
    '.history',
    '.sh_history',
    '.bash_history',
    '.zsh_history',
    '.gnupg',
    '.ssh',
    '.kamakirc',
    '.kamaki.history',
    '.kamaki.log'
]


def check_sensitive_userdata(value):
    "Do not allow a zero string"

    if len(value) == 0:
        raise ValueError("Filename cannot be empty")

    return value


class Unix(OSBase):
    """OS class for Unix"""
    @add_sysprep_param(
        'sensitive_userdata', 'list:string', SENSITIVE_USERDATA,
        'Files in the home directory of each user that should be removed',
        check=check_sensitive_userdata)
    def __init__(self, image, **kwargs):
        super(Unix, self).__init__(image, **kwargs)

    def _mountpoints(self):
        """Return mountpoints in the correct order.
        / should be mounted before /boot or /usr, /usr befor /usr/bin ...
        """
        mps = self.image.g.inspect_get_mountpoints(self.root)

        def compare(a, b):
            if len(a[0]) > len(b[0]):
                return 1
            elif len(a[0]) == len(b[0]):
                return 0
            else:
                return -1
        mps.sort(compare)

        for mp in mps:
            yield mp

    def _do_mount(self, readonly):
        """Mount partitions in the correct order"""

        critical_mpoints = ('/', '/etc', '/root', '/home', '/var')

        mopts = 'ro' if readonly else 'rw'
        for mp, dev in self._mountpoints():
            try:
                self.image.g.mount_options(mopts, dev, mp)
            except RuntimeError as msg:
                if mp in critical_mpoints:
                    self._mount_error = str(msg)
                    return False
                else:
                    self._mount_warnings.append('%s (ignored)' % msg)

        return True

    def _do_collect_metadata(self):
        super(Unix, self)._do_collect_metadata()

        bin_prefixes = ('', '/usr', '/usr/local')
        gui = False

        paths = ['%s/bin/%s' % (p, X11_EXECUTABLE) for p in bin_prefixes]
        for path in paths:
            if self.image.g.is_file(path):
                gui = True
                break

        if not gui:
            self.meta['GUI'] = "No GUI"
        else:
            self.meta['GUI'] = 'Unknown'

        desktop = []
        for exe, session in DESKTOPSESSIONS.items():
            paths = ["%s/bin/%s" % (p, e) for p in bin_prefixes for e in exe]
            for path in paths:
                if self.image.g.is_file(path):
                    desktop.append(session)
                    break
        if gui and len(desktop) != 0:
            self.meta['GUI'] = " | ".join(desktop)

    def ssh_connection_options(self, users):
        """Returns a list of valid ssh connection options"""

        def sshd_config():
            """Read /etc/ssh/sshd_config and return it as a dictionary"""
            config = {}
            fname = '/etc/ssh/sshd_config'

            if not self.image.g.is_file(fname):
                return {}

            for line in self.image.g.cat(fname).splitlines():
                line = line.split('#')[0].strip()
                if not len(line):
                    continue
                line = line.split()
                config[line[0]] = line[1:]
            return config

        config = sshd_config()
        try:
            port = int(config['Port'][0])
        except:
            port = 22

        if 'PermitRootLogin' in config and config['PermitRootLogin'] == 'no':
            if 'root' in users:
                users.remove('root')

        return {'port': port, 'users': users}

    @sysprep('Removing files under /var/cache')
    def _cleanup_cache(self):
        """Remove all regular files under /var/cache"""

        self._foreach_file('/var/cache', self.image.g.rm, ftype='r')

    @sysprep('Removing files under /tmp and /var/tmp')
    def _cleanup_tmp(self):
        """Remove all files under /tmp and /var/tmp"""

        self._foreach_file('/tmp', self.image.g.rm_rf, maxdepth=1)
        self._foreach_file('/var/tmp', self.image.g.rm_rf, maxdepth=1)

    @sysprep('Emptying all files under /var/log')
    def _cleanup_log(self):
        """Empty all files under /var/log"""

        self._foreach_file('/var/log', self.image.g.truncate, ftype='r')

    @sysprep('Removing files under /var/mail & /var/spool/mail', enabled=False)
    def _cleanup_mail(self):
        """Remove all files under /var/mail and /var/spool/mail"""

        self._foreach_file('/var/spool/mail', self.image.g.rm_rf, maxdepth=1)

        self._foreach_file('/var/mail', self.image.g.rm_rf, maxdepth=1)

    @sysprep('Removing sensitive user data')
    def _cleanup_userdata(self):
        """Delete sensitive user data"""

        homedirs = ['/root']
        if self.image.g.is_dir('/home/'):
            homedirs += self._ls('/home/')

        action = self.image.g.rm_rf
        if self._scrub_support:
            action = self.image.g.scrub_file
        else:
            self.out.warn("Sensitive data won't be scrubbed (not supported)")

        sensitive_userdata = self.sysprep_params['sensitive_userdata'].value
        for homedir in homedirs:
            for data in sensitive_userdata:
                fname = "%s/%s" % (homedir, data)
                if self.image.g.is_file(fname):
                    action(fname)
                elif self.image.g.is_dir(fname):
                    self._foreach_file(fname, action, ftype='r')

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2016 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module hosts the code that performs the host bundling operation. By
using the create_image method of the BundleVolume class the user can create an
image out of the running system.
"""

import os
import re
import tempfile
import uuid
from collections import namedtuple

import parted

from image_creator.rsync import Rsync
from image_creator.util import get_command
from image_creator.util import FatalError
from image_creator.util import try_fail_repeat
from image_creator.util import free_space
from image_creator.gpt import GPTPartitionTable

findfs = get_command('findfs')
dd = get_command('dd')
dmsetup = get_command('dmsetup')
losetup = get_command('losetup')
mount = get_command('mount')
umount = get_command('umount')
blkid = get_command('blkid')
tune2fs = get_command('tune2fs')

MKFS_OPTS = {'ext2': {'force': '-F', 'uuid': '-U', 'label': '-L'},
             'ext3': {'force': '-F', 'uuid': '-U', 'label': '-L'},
             'ext4': {'force': '-F', 'uuid': '-U', 'label': '-L'},
             'reiserfs': {'force': '-ff', 'uuid': '-u', 'label': '-l'},
             'btrfs':  {'force': '-f', 'label': '-L'},
             'minix': {},
             'xfs': {'force': '-f', 'label': '-L'},
             'jfs': {'force': '-f', 'label': '-L'},
             'ntfs': {'force': '-F', 'label': '-L'},
             'msdos': {'uuid': '-i'},
             'vfat': {'uuid': '-i'}}

UUID_UPDATE = {
    'ext2': lambda d, u: tune2fs('-U', u, d),
    'ext3': lambda d, u: tune2fs('-U', u, d),
    'ext4': lambda d, u: tune2fs('-U', u, d),
    'reiserfs': lambda d, u: get_command('reiserfstune')('-u', u, d),
    'xfs': lambda d, u: get_command('xfs_admin')('-U', u, d),
    'jfs': lambda d, u: get_command('jfstune')('-U', u, d),
    'ntfs': lambda d, u: get_command('ntfslable')('--new-serial=%s' % u, d)}


def mkfs(fs, device, uuid=None, label=None):
    """Create a filesystem on the device"""

    mkfs = get_command('mkfs.%s' % fs)

    args = []

    if 'force' in MKFS_OPTS[fs]:
        args.append(MKFS_OPTS[fs]['force'])

    if label:
        args.append(MKFS_OPTS[fs]['label'])
        args.append(label)

    if 'uuid' in MKFS_OPTS[fs] and uuid:
        args.append(MKFS_OPTS[fs]['uuid'])
        args.append(uuid)

    args.append(device)

    mkfs(*args)

    if 'uuid' not in MKFS_OPTS[fs] and 'uuid':
        UUID_UPDATE[fs](device, uuid)


def read_fstable(f):
    """Use this generator to iterate over the lines of an fstab file"""

    if not os.path.isfile(f):
        raise FatalError("Unable to open: `%s'. File is missing." % f)

    FileSystemTableEntry = namedtuple('FileSystemTableEntry',
                                      'dev mpoint fs opts freq passno')
    with open(f) as table:
        for line in iter(table):
            entry = line.split('#')[0].strip().split()
            if len(entry) != 6:
                continue
            yield FileSystemTableEntry(*entry)


def get_root_partition():
    """Return the fstab entry associated with the root file system"""
    for entry in read_fstable('/etc/fstab'):
        if entry.mpoint == '/':
            return entry.dev

    raise FatalError("Unable to find root device in /etc/fstab")


def is_mpoint(path):
    """Check if a directory is currently a mount point"""
    for entry in read_fstable('/proc/mounts'):
        if entry.mpoint == path:
            return True
    return False


def get_mount_options(device):
    """Return the mount entry associated with a mounted device"""
    for entry in read_fstable('/proc/mounts'):
        if not entry.dev.startswith('/'):
            continue

        if os.path.realpath(entry.dev) == os.path.realpath(device):
            return entry

    return None


def get_partitions(disk):
    """Returns a list with the partitions of the provided disk"""
    Partition = namedtuple('Partition', 'num start end type fs')

    partitions = []
    for p in disk.partitions:
        num = p.number
        start = p.geometry.start
        end = p.geometry.end
        ptype = p.type
        fs = p.fileSystem.type if p.fileSystem is not None else ''
        partitions.append(Partition(num, start, end, ptype, fs))

    return partitions


def map_partition(dev, num, start, end):
    """Map a partition into a block device using the device mapper"""
    name = os.path.basename(dev) + "_" + uuid.uuid4().hex
    tablefd, table = tempfile.mkstemp()
    try:
        try:
            size = end - start + 1
            os.write(tablefd, "0 %d linear %s %d" % (size, dev, start))
        finally:
            os.close(tablefd)
        dmsetup('create', "%sp%d" % (name, num), table)
    finally:
        os.unlink(table)

    return "/dev/mapper/%sp%d" % (name, num)


def unmap_partition(dev):
    """Unmap a previously mapped partition"""
    if not os.path.exists(dev):
        return

    try_fail_repeat(dmsetup, 'remove', dev.split('/dev/mapper/')[1])


def mount_all(target, devs):
    """Mount a list of file systems in mount points relative to target"""
    devs.sort(key=lambda d: d[1])
    for dev, mpoint, options in devs:
        absmpoint = os.path.abspath(target + mpoint)
        if not os.path.exists(absmpoint):
            os.makedirs(absmpoint)

        if len(options) > 0:
            mount(dev, absmpoint, '-o', ",".join(options))
        else:
            mount(dev, absmpoint)


def umount_all(target):
    """Umount all file systems that are mounted under the target directory"""
    mpoints = []
    for entry in read_fstable('/proc/mounts'):
        if entry.mpoint.startswith(os.path.abspath(target)):
            mpoints.append(entry.mpoint)

    mpoints.sort()
    for mpoint in reversed(mpoints):
        try_fail_repeat(umount, mpoint)


class BundleVolume(object):
    """This class can be used to create an image out of the running system"""

    def __init__(self, out, meta, tmp=None):
        """Create an instance of the BundleVolume class."""
        self.out = out
        self.meta = meta
        self.tmp = tmp

        self.out.info('Searching for root device ...', False)
        root = get_root_partition()

        if root.startswith("UUID=") or root.startswith("LABEL="):
            root = findfs(root).stdout.strip()

        if not re.match('/dev/x?[hsv]d[a-z][1-9]*$', root):
            raise FatalError("Don't know how to handle root device: %s" % root)

        out.success(root)

        disk_file = re.split('[0-9]', root)[0]
        device = parted.Device(disk_file)
        self.disk = parted.Disk(device)

    def _create_partition_table(self, image):
        """Copy the partition table of the host system into the image"""

        # Copy the MBR and the space between the MBR and the first partition.
        # In MSDOS partition tables GRUB Stage 1.5 is located there.
        # In GUID partition tables the Primary GPT Header is there.
        first_sector = self.disk.getPrimaryPartitions()[0].geometry.start

        dd('if=%s' % self.disk.device.path, 'of=%s' % image,
           'bs=%d' % self.disk.device.sectorSize,
           'count=%d' % first_sector, 'conv=notrunc')

        if self.disk.type == 'gpt':
            # Copy the Secondary GPT Header
            table = GPTPartitionTable(self.disk.device.path)
            dd('if=%s' % self.disk.device.path, 'of=%s' % image,
               'bs=%d' % self.disk.device.sectorSize, 'conv=notrunc',
               'seek=%d' % table.primary.last_usable_lba,
               'skip=%d' % table.primary.last_usable_lba)

        # Create the Extended boot records (EBRs) in the image
        extended = self.disk.getExtendedPartition()
        if not extended:
            return

        # Extended boot records precede the logical partitions they describe
        logical = self.disk.getLogicalPartitions()
        start = extended.geometry.start
        for i in range(len(logical)):
            end = logical[i].geometry.start - 1
            dd('if=%s' % self.disk.device.path, 'of=%s' % image,
               'count=%d' % (end - start + 1), 'conv=notrunc',
               'seek=%d' % start, 'skip=%d' % start)
            start = logical[i].geometry.end + 1

    def _shrink_partitions(self, image):
        """Remove the last partition of the image if it is a swap partition and
        shrink the partition before that. Make sure it can still host all the
        files the corresponding host file system hosts
        """
        image_disk = parted.Disk(parted.Device(image))

        def is_extended(partition):
            """Returns True if the partition is extended"""
            return partition.type == parted.PARTITION_EXTENDED

        def is_logical(partition):
            """Returns True if the partition is logical"""
            return partition.type == parted.PARTITION_LOGICAL

        partitions = get_partitions(self.disk)

        last = partitions[-1]
        new_end = last.end
        if last.fs == 'linux-swap(v1)':
            MB = 2 ** 20
            size = (last.end - last.start + 1) * self.disk.device.sectorSize
            self.meta['SWAP'] = "%d:%s" % (last.num, (size + MB - 1) // MB)

            image_disk.deletePartition(
                image_disk.getPartitionBySector(last.start))
            image_disk.commitToDevice()

            if is_logical(last) and last.num == 5:
                extended = image_disk.getExtendedPartition()
                image_disk.deletePartition(extended)
                image_disk.commitToDevice()
                partitions.remove(filter(is_extended, partitions)[0])

            partitions.remove(last)
            last = partitions[-1]

            new_end = last.end

        mount_options = get_mount_options(
            self.disk.getPartitionBySector(last.start).path)
        if mount_options is not None:
            stat = os.statvfs(mount_options.mpoint)
            # Shrink the last partition. The new size should be the size of the
            # occupied blocks
            blcks = stat.f_blocks - stat.f_bavail
            new_size = (blcks * stat.f_frsize) // self.disk.device.sectorSize

            # Add 10% just to be on the safe side
            part_end = last.start + (new_size * 11) // 10
            # Align to 2048
            part_end = ((part_end + 2047) // 2048) * 2048

            # Make sure the partition starts where the old partition started.
            constraint = parted.Constraint(device=image_disk.device)
            constraint.startRange = parted.Geometry(device=image_disk.device,
                                                    start=last.start, length=1)

            image_disk.setPartitionGeometry(
                image_disk.getPartitionBySector(last.start), constraint,
                start=last.start, end=part_end)
            image_disk.commitToDevice()

            # Parted may have changed this for better alignment
            part_end = image_disk.getPartitionBySector(last.start).geometry.end
            last = last._replace(end=part_end)
            partitions[-1] = last

            new_end = part_end

            if last.type == parted.PARTITION_LOGICAL:
                # Fix the extended partition
                image_disk.minimizeExtendedPartition()

        return (new_end, get_partitions(image_disk))

    def _to_exclude(self):
        """Find which directories to exclude during the image copy. This is
        accomplished by checking which directories serve as mount points for
        virtual file systems
        """
        excluded = ['/tmp', '/var/tmp']
        if self.tmp is not None:
            excluded.append(self.tmp)
        local_filesystems = MKFS_OPTS.keys() + ['rootfs']
        for entry in read_fstable('/proc/mounts'):
            if entry.fs in local_filesystems:
                continue

            mpoint = entry.mpoint
            if mpoint in excluded:
                continue

            descendants = [e for e in excluded if e.startswith(mpoint + '/')]
            if len(descendants):
                for d in descendants:
                    excluded.remove(d)
                excluded.append(mpoint)
                continue

            dirname = mpoint
            found_ancestor = False
            while dirname != '/':
                (dirname, _) = os.path.split(dirname)
                if dirname in excluded:
                    found_ancestor = True
                    break

            if not found_ancestor:
                excluded.append(mpoint)

        return excluded

    def _create_filesystems(self, image, partitions):
        """Fill the image with data. Host file systems that are not currently
        mounted are binary copied into the image. For mounted file systems, a
        file system level copy is performed.
        """

        filesystem = {}
        orig_dev = {}
        for p in self.disk.partitions:
            filesystem[p.number] = get_mount_options(p.path)
            orig_dev[p.number] = p.path

        unmounted = [p for p in partitions if filesystem[p.num] is None]
        mounted = [p for p in partitions if filesystem[p.num] is not None]

        # For partitions that are not mounted right now, we can simply dd them
        # into the image.
        for p in unmounted:
            self.out.info('Cloning partition %d ... ' % p.num, False)
            dd('if=%s' % self.disk.device.path, 'of=%s' % image,
               'count=%d' % (p.end - p.start + 1), 'conv=notrunc',
               'seek=%d' % p.start, 'skip=%d' % p.start)
            self.out.success("done")

        loop = str(losetup('-f', '--show', image)).strip()

        # Recreate mounted file systems
        mapped = {}
        try:
            for p in mounted:
                i = p.num
                mapped[i] = map_partition(loop, i, p.start, p.end)

            new_uuid = {}
            # Create the file systems
            for i, dev in mapped.iteritems():
                uuid = blkid(
                    '-s', 'UUID', '-o', 'value', orig_dev[i]).stdout.strip()
                label = blkid(
                    '-s', 'LABEL', '-o', 'value', orig_dev[i]).stdout.strip()
                fs = filesystem[i].fs

                self.out.info('Creating %s file system on partition %d ... '
                              % (fs, i), False)
                mkfs(fs, dev, uuid=uuid, label=label)

                # For ext[234] enable the default mount options
                if re.match('^ext[234]$', fs):
                    mopts = filter(
                        lambda p: p.startswith('Default mount options:'),
                        tune2fs('-l', orig_dev[i]).splitlines()
                    )[0].split(':')[1].strip().split()

                    if not (len(mopts) == 1 and mopts[0] == '(none)'):
                        for opt in mopts:
                            tune2fs('-o', opt, dev)

                self.out.success('done')
                new_uuid[i] = blkid(
                    '-s', 'UUID', '-o', 'value', dev).stdout.strip()

            target = tempfile.mkdtemp()
            devs = []
            for i in mapped.keys():
                fs = filesystem[i].fs
                mpoint = filesystem[i].mpoint
                opts = []
                for opt in filesystem[i].opts.split(','):
                    if opt in ('acl', 'user_xattr'):
                        opts.append(opt)
                devs.append((mapped[i], mpoint, opts))
            try:
                mount_all(target, devs)

                excluded = self._to_exclude()

                rsync = Rsync(self.out)

                for excl in excluded + [image]:
                    rsync.exclude(excl)

                rsync.archive().hard_links().xattrs().sparse().acls()
                rsync.run('/', target, 'host', 'temporary image')

                # Create missing mount points. We cannot determine the
                # ownership and the mode of the real directory. Make them
                # inherit those properties from their parent directory.
                for excl in excluded:
                    dirname = os.path.dirname(excl)
                    stat = os.stat(dirname)
                    os.mkdir(target + excl)
                    os.chmod(target + excl, stat.st_mode)
                    os.chown(target + excl, stat.st_uid, stat.st_gid)

                # /tmp and /var/tmp are special cases. We exclude then even if
                # they aren't mount points. Restore their permissions.
                for excl in ('/tmp', '/var/tmp'):
                    if is_mpoint(excl):
                        os.chmod(target + excl, 041777)
                        os.chown(target + excl, 0, 0)
                    else:
                        stat = os.stat(excl)
                        os.chmod(target + excl, stat.st_mode)
                        os.chown(target + excl, stat.st_uid, stat.st_gid)

            finally:
                umount_all(target)
                os.rmdir(target)
        finally:
            for dev in mapped.values():
                unmap_partition(dev)
            losetup('-d', loop)

    def create_image(self, image):
        """Given an image filename, this method will create an image out of the
        running system.
        """

        size = self.disk.device.length * self.disk.device.sectorSize

        # Create sparse file to host the image
        fd = os.open(image, os.O_WRONLY | os.O_CREAT)
        try:
            os.ftruncate(fd, size)
        finally:
            os.close(fd)

        self._create_partition_table(image)
        end_sector, partitions = self._shrink_partitions(image)

        if self.disk.type == 'gpt':
            old_size = size
            size = (end_sector + 1) * self.disk.device.sectorSize
            ptable = GPTPartitionTable(image)
            size = ptable.shrink(size, old_size)
        else:
            # Align to 2048
            end_sector = ((end_sector + 2047) // 2048) * 2048
            size = (end_sector + 1) * self.disk.device.sectorSize

        # Truncate image to the new size.
        fd = os.open(image, os.O_RDWR)
        try:
            os.ftruncate(fd, size)
        finally:
            os.close(fd)

        # Check if the available space is enough to host the image
        dirname = os.path.dirname(image)
        self.out.info("Examining available space ...", False)
        if free_space(dirname) <= size:
            raise FatalError("Not enough space under %s to host the temporary "
                             "image" % dirname)
        self.out.success("sufficient")

        self._create_filesystems(image, partitions)

        return image

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2015 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This package provides various classes for preparing different Operating
Systems for image creation.
"""

from image_creator.util import FatalError
from image_creator.bootloader import mbr_bootinfo, vbr_bootinfo

import textwrap
import re
from collections import namedtuple
from functools import wraps

OSTYPE_ORDER = {
    "windows": 8,
    "linux": 7,
    "freebsd": 6,
    "netbsd": 5,
    "openbsd": 4,
    "hurd": 3,
    "dos": 2,
    "minix": 1
}


def os_cls(distro, osfamily):
    """Given the distro name and the osfamily, return the appropriate OSBase
    derived class
    """

    # hyphens are not allowed in module names
    distro = distro.replace('-', '_').lower()
    osfamily = osfamily.replace('-', '_').lower()

    if distro == 'unknown':
        distro = osfamily

    try:
        module = __import__("image_creator.os_type.%s" % distro,
                            fromlist=['image_creator.os_type'])
        classname = distro.capitalize()
    except ImportError:
        try:
            module = __import__("image_creator.os_type.%s" % osfamily,
                                fromlist=['image_creator.os_type'])
            classname = osfamily.capitalize()
        except ImportError:
            raise FatalError("Unknown OS name: `%s'" % osfamily)

    return getattr(module, classname)


def add_prefix(target):
    """Decorator that adds a prefix to the result of a function"""
    def wrapper(self, *args):
        prefix = args[0]
        return [prefix + path for path in target(self, *args)]
    return wrapper


def sysprep(message, enabled=True, **kwargs):
    """Decorator for system preparation tasks"""
    def wrapper(method):
        assert method.__name__.startswith('_'), \
            "Invalid sysprep name:` %s'. Should start with _" % method.__name__

        method._sysprep = True
        method._sysprep_enabled = enabled
        method._sysprep_nomount = False

        for key, val in kwargs.items():
            setattr(method, "_sysprep_%s" % key, val)

        @wraps(method)
        def inner(self, print_message=True):
            if print_message:
                self.out.info(message % self.sysprep_params)
            return method(self)

        return inner
    return wrapper


class SysprepParam(object):
    """This class represents a system preparation parameter"""

    def __init__(self, name, type, default, description, **kwargs):

        self.name = name
        self.is_list = type.startswith('list:')
        self.type = type.split(':', 1)[1] if self.is_list else type
        self.default = default
        self.description = description
        self.value = default
        self.error = None
        self.check = kwargs['check'] if 'check' in kwargs else lambda x: x
        self.hidden = kwargs['hidden'] if 'hidden' in kwargs else False

        assert hasattr(self, "_check_%s" % self.type), \
            "Invalid type: %s" % self.type

    def __str__(self):
        """Return the value as a string"""
        return str(self.value)

    def set_value(self, value):
        """Update the value of the parameter"""

        check_type = getattr(self, "_check_%s" % self.type)

        tmp = []

        for item in value if self.is_list else [value]:
            try:
                tmp.append(self.check(check_type(item)))
            except ValueError as e:
                self.error = e.message
                return False

        self.value = tmp if self.is_list else tmp[0]

        return True

    @staticmethod
    def _check_posint(value):
        """Check if the value is a positive integer"""
        try:
            value = int(value)
        except ValueError:
            raise ValueError("Invalid number")

        if value <= 0:
            raise ValueError("Value is negative or zero")

        return value

    @staticmethod
    def _check_string(value):
        """Check if a value is a string"""
        return str(value)

    @staticmethod
    def _check_file(value):
        """Check if the value is a valid filename"""

        value = str(value)
        if len(value) == 0:
            return ""

        import os

        def isblockdev(filename):
            import stat
            try:
                return stat.S_ISBLK(os.stat(filename).st_mode)
            except OSError:
                return False
        if os.path.isfile(value) or isblockdev(value):
            return value

        raise ValueError("Invalid filename")

    @staticmethod
    def _check_dir(value):
        """Check if the value is a valid directory"""

        value = str(value)
        if len(value) == 0:
            return ""

        import os
        if os.path.isdir(value):
            return value

        raise ValueError("Invalid dirname")


def add_sysprep_param(name, type, default, descr, **kwargs):
    """Decorator for __init__ that adds the definition for a system preparation
    parameter in an instance of an os_type class
    """
    extra = kwargs

    def wrapper(init):
        @wraps(init)
        def inner(self, *args, **kwargs):

            if not hasattr(self, 'sysprep_params'):
                self.sysprep_params = {}

            self.sysprep_params[name] = \
                SysprepParam(name, type, default, descr, **extra)
            init(self, *args, **kwargs)
        return inner
    return wrapper


def del_sysprep_param(name):
    """Decorator for __init__ that deletes a previously added sysprep parameter
    definition from an instance of a os_type class.
    """
    def wrapper(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            del self.sysprep_params[name]
            func(self, *args, **kwargs)
        return inner
    return wrapper


class OSBase(object):
    """Basic operating system class"""

    def __init__(self, image, **kwargs):
        self.image = image

        self.root = image.root
        self.out = image.out

        # Could be defined in a decorator
        if not hasattr(self, 'sysprep_params'):
            self.sysprep_params = {}

        if 'sysprep_params' in kwargs:
            for key, val in kwargs['sysprep_params'].items():
                if key not in self.sysprep_params:
                    self.out.warn("Ignoring invalid `%s' parameter." % key)
                    continue
                param = self.sysprep_params[key]
                if param.is_list:
                    def split_in_comma(val):
                        tmp = val.split(',')
                        prev = ""
                        for i in xrange(len(tmp)):
                            item = prev + tmp[i]
                            if item.endswith('\\'):
                                prev = item[:-1]
                                continue
                            prev = ""
                            yield item
                    val = list(split_in_comma(val))

                if not param.set_value(val):
                    raise FatalError("Invalid value for sysprep parameter: "
                                     "`%s'. Reason: %s" % (key, param.error))

        self.meta = self.image.meta
        self.shrinked = False

        # This will host the error if mount fails
        self._mount_error = ""
        self._mount_warnings = []
        self._mounted = None

        # Many guestfs compilations don't support scrub
        self._scrub_support = True
        try:
            self.image.g.available(['scrub'])
        except RuntimeError:
            self._scrub_support = False

        # Create a list of available syspreps
        self._sysprep_tasks = {}
        for name in dir(self):
            obj = getattr(self, name)
            if not hasattr(obj, '_sysprep'):
                continue
            self._sysprep_tasks[name] = obj._sysprep_enabled

        self._cleanup_jobs = {}

    def _add_cleanup(self, namespace, job, *args):
        """Add a new job in a cleanup list"""

        if namespace not in self._cleanup_jobs:
            self._cleanup_jobs[namespace] = []

        self._cleanup_jobs[namespace].append((job, args))

    def _cleanup(self, namespace):
        """Run the cleanup tasks that are defined under a specific namespace"""

        if namespace not in self._cleanup_jobs:
            self.out.warn("Cleanup namespace: `%s' is not defined", namespace)
            return

        while len(self._cleanup_jobs[namespace]):
            job, args = self._cleanup_jobs[namespace].pop()
            job(*args)

        del self._cleanup_jobs[namespace]

    def inspect(self):
        """Inspect the media to check if it is supported"""

        if self.image.is_unsupported():
            return

        self.out.info('Running OS inspection:')
        with self.mount(readonly=True, silent=True):
            self._do_inspect()
        self.out.info()

    def collect_metadata(self):
        """Collect metadata about the OS"""

        self.out.info('Collecting image metadata ...', False)

        mbr = self.image.g.pread_device('/dev/sda', 512, 0)
        self.meta['BOOTSTRAP'] = mbr_bootinfo(mbr)

        with self.mount(readonly=True, silent=True):
            self._do_collect_metadata()

        self.out.success('done')
        self.out.info()

    def list_syspreps(self):
        """Returns a list of sysprep objects"""
        return [getattr(self, name) for name in self._sysprep_tasks]

    def sysprep_info(self, obj):
        """Returns information about a sysprep object"""
        assert hasattr(obj, '_sysprep'), "Object is not a sysprep"

        SysprepInfo = namedtuple("SysprepInfo", "name description display")

        name = obj.__name__.replace('_', '-')[1:]
        description = textwrap.dedent(obj.__doc__)
        display = getattr(obj, '_sysprep_display',
                          name.replace('-', ' ').capitalize())

        return SysprepInfo(name, description, display)

    def get_sysprep_by_name(self, name):
        """Returns the sysprep object with the given name"""

        method_name = '_' + name.replace('-', '_')

        if hasattr(self, method_name):
            method = getattr(self, method_name)

            if hasattr(method, '_sysprep'):
                return method

        return None

    def enable_sysprep(self, obj):
        """Enable a system preparation operation"""
        assert hasattr(obj, '_sysprep'), "Object is not a sysprep"
        assert obj.__name__ in self._sysprep_tasks, "Sysprep already executed"

        self._sysprep_tasks[obj.__name__] = True

    def disable_sysprep(self, obj):
        """Disable a system preparation operation"""
        assert hasattr(obj, '_sysprep'), "Object is not a sysprep"
        assert obj.__name__ in self._sysprep_tasks, "Sysprep already executed"

        self._sysprep_tasks[obj.__name__] = False

    def sysprep_enabled(self, obj):
        """Returns True if this system praparation operation is enabled"""
        assert hasattr(obj, '_sysprep'), "Object is not a sysprep"
        assert obj.__name__ in self._sysprep_tasks, "Sysprep already executed"

        return self._sysprep_tasks[obj.__name__]

    def print_metadata(self):
        """Print the image metadata"""

        self.out.info("Detected image metadata:")

        col_width = max(len(key) for key in self.meta) + 2
        for key, val in self.meta.items():
            self.out.info("%s %s" % (key.ljust(col_width), val))

    def print_syspreps(self):
        """Print enabled and disabled system preparation operations"""

        syspreps = self.list_syspreps()
        enabled = [s for s in syspreps if self.sysprep_enabled(s)]
        disabled = [s for s in syspreps if not self.sysprep_enabled(s)]

        wrapper = textwrap.TextWrapper()
        wrapper.subsequent_indent = '\t'
        wrapper.initial_indent = '\t'
        wrapper.width = 72

        self.out.info("Enabled system preparation operations:")
        if len(enabled) == 0:
            self.out.info("(none)")
        else:
            for sysprep in enabled:
                name = sysprep.__name__.replace('_', '-')[1:]
                descr = wrapper.fill(textwrap.dedent(sysprep.__doc__))
                self.out.info('    %s:\n%s\n' % (name, descr))

        self.out.info("Disabled system preparation operations:")
        if len(disabled) == 0:
            self.out.info("(none)")
        else:
            for sysprep in disabled:
                name = sysprep.__name__.replace('_', '-')[1:]
                descr = wrapper.fill(textwrap.dedent(sysprep.__doc__))
                self.out.info('    %s:\n%s\n' % (name, descr))

    def print_sysprep_params(self):
        """Print the system preparation parameter the user may use"""

        self.out.info("System preparation parameters:")
        self.out.info()

        public_params = [(n, p) for n, p in self.sysprep_params.items()
                         if not p.hidden]
        if len(public_params) == 0:
            self.out.info("(none)")
            return

        wrapper = textwrap.TextWrapper()
        wrapper.subsequent_indent = "             "
        wrapper.width = 80

        for name, param in public_params:
            if param.hidden:
                continue
            self.out.info("NAME:".ljust(13) + name)
            self.out.info(wrapper.fill("DESCRIPTION:".ljust(13) +
                                       "%s" % param.description))
            self.out.info("TYPE:".ljust(13) + "%s%s" %
                          ("list:" if param.is_list else "", param.type))
            self.out.info("VALUE:".ljust(13) + "%s" %
                          ("\n".ljust(14).join(param.value) if param.is_list
                           else param.value))
            self.out.info()

    def do_sysprep(self):
        """Prepare system for image creation."""

        self.out.info('Preparing system for image creation:')

        if self.image.is_unsupported():
            self.out.warn(
                "System preparation is disabled for unsupported media")
            return

        enabled = [s for s in self.list_syspreps() if self.sysprep_enabled(s)]
        size = len(enabled)
        cnt = 0

        def exec_sysprep(cnt, size, task):
            self.out.info(('(%d/%d)' % (cnt, size)).ljust(7), False)
            task()
            del self._sysprep_tasks[task.__name__]

        with self.mount():
            for task in [t for t in enabled if t._sysprep_nomount is False]:
                cnt += 1
                exec_sysprep(cnt, size, task)

        for task in [t for t in enabled if t._sysprep_nomount]:
            cnt += 1
            exec_sysprep(cnt, size, task)

        self.out.info()

    @sysprep('Shrinking image (may take a while)', nomount=True)
    def _shrink(self):
        """Shrink the last file system and update the partition table"""
        device = self.image.shrink()
        self.shrinked = True

        # Check the Volume Boot Record of the shrinked partition to determine
        # if a bootloader is present on it.
        vbr = self.image.g.pread_device(device, 512, 0)
        bootloader = vbr_bootinfo(vbr)

        if bootloader == 'syslinux':
            # EXTLINUX needs to be reinstalled after shrinking
            with self.mount(silent=True):
                self.out.info("Reinstalling extlinux ...", False)
                if self.image.g.is_dir('/boot/extlinux'):
                    extdir = '/boot/extlinux'
                elif self.image.g.is_dir('/boot/syslinux'):
                    extdir = '/boot/syslinux'
                else:
                    extdir = '/boot'

                self.image.g.command(['extlinux', '--install', extdir])
                self.out.success("done")

    @property
    def ismounted(self):
        return self._mounted is not None

    def mount(self, readonly=False, silent=False, fatal=True):
        """Returns a context manager for mounting an image"""

        if not silent:
            output = self.out.info
            success = self.out.success
            warn = self.out.warn
        else:
            def output(msg='', nl=True):
                pass
            success = warn = output

        mount_type = 'read-only' if readonly else 'read-write'
        output("Mounting the media %s ..." % mount_type, False)

        self._mount_error = ""
        del self._mount_warnings[:]

        try:
            mounted = self._do_mount(readonly)
        except:
            self.image.g.umount_all()
            raise

        if not mounted:
            msg = "Unable to mount the media %s. Reason: %s" % \
                (mount_type, self._mount_error)
            if fatal:
                raise FatalError(msg)
            else:
                warn(msg)

        for warning in self._mount_warnings:
            warn(warning)

        if mounted:
            success('done')

        parent = self

        class Mount(object):
            """The Mount context manager"""
            def __enter__(self):
                pass

            def __exit__(self, exc_type, exc_value, traceback):
                self.umount()

            def umount(self):
                output("Umounting the media ...", False)
                parent.image.g.umount_all()
                parent._mounted = None
                success('done')

        self._mounted = Mount()
        return self._mounted

    def umount(self):
        """Umount a previously mounted image"""
        if self._mounted is None:
            self.out.warn("Ignoring the umount request.")
            return

        self._mounted.umount()

    def check_version(self, major, minor):
        """Checks the OS version against the one specified by the major, minor
        tuple.

        Returns:
            < 0 if the OS version is smaller than the specified one
            = 0 if they are equal
            > 0 if it is greater
        """
        guestfs = self.image.g
        for a, b in ((guestfs.inspect_get_major_version(self.root), major),
                     (guestfs.inspect_get_minor_version(self.root), minor)):
            if a != b:
                return a - b

        return 0

    @add_prefix
    def _ls(self, directory):
        """List the name of all files under a directory"""
        return self.image.g.ls(directory)

    @add_prefix
    def _find(self, directory):
        """List the name of all files recursively under a directory"""
        return self.image.g.find(directory)

    def _foreach_file(self, directory, action, **kwargs):
        """Perform an action recursively on all files under a directory.

        The following options are allowed:

        * maxdepth: If defined, the action will not be performed on files that
          are below this level of directories under the directory parameter.

        * ftype: The action will only be performed on files of this type. For a
          list of all allowed file types, see here:
          http://libguestfs.org/guestfs.3.html#guestfs_readdir

        * exclude: Exclude all files that follow this pattern.

        * include: Only include files that follow this pattern.
        """
        if not self.image.g.is_dir(directory):
            self.out.warn("Directory: `%s' does not exist!" % directory)
            return

        maxdepth = None if 'maxdepth' not in kwargs else kwargs['maxdepth']
        if maxdepth == 0:
            return

        # maxdepth -= 1
        maxdepth = None if maxdepth is None else maxdepth - 1
        kwargs['maxdepth'] = maxdepth

        exclude = None if 'exclude' not in kwargs else kwargs['exclude']
        include = None if 'include' not in kwargs else kwargs['include']
        ftype = None if 'ftype' not in kwargs else kwargs['ftype']

        def has_ftype(f, ftype):
            """Returns True if the type of the file is ftype"""
            return ftype is None and True or f['ftyp'] == ftype

        for f in self.image.g.readdir(directory):
            if f['name'] in ('.', '..'):
                continue

            full_path = "%s/%s" % (directory, f['name'])

            if exclude and re.match(exclude, full_path):
                continue

            if include and not re.match(include, full_path):
                continue

            if has_ftype(f, 'd'):
                self._foreach_file(full_path, action, **kwargs)

            if has_ftype(f, ftype):
                action(full_path)

    def _do_inspect(self):
        """helper method for inspect"""
        self.out.warn("No inspection method available")

    def _do_collect_metadata(self):
        """helper method for collect_metadata"""

        try:
            self.meta['ROOT_PARTITION'] = \
                "%d" % self.image.g.part_to_partnum(self.root)
        except RuntimeError:
            self.out.warn("Unable to identify the partition number from root "
                          "partition: %s" % self.root)

        osfamily = self.image.g.inspect_get_type(self.root)
        distro = self.image.g.inspect_get_distro(self.root)
        name = self.image.g.inspect_get_product_name(self.root)

        self.meta['OSFAMILY'] = osfamily
        self.meta['OS'] = distro if distro != "unknown" else osfamily
        self.meta['DESCRIPTION'] = name
        try:
            self.meta['SORTORDER'] = 1000000 * OSTYPE_ORDER[osfamily]
        except KeyError:
            self.meta['SORTORDER'] = 0

    def _do_mount(self, readonly):
        """helper method for mount"""
        try:
            self.image.g.mount_options(
                'ro' if readonly else 'rw', self.root, '/')
        except RuntimeError as msg:
            self._mount_error = str(msg)
            return False

        return True

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2015 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Normal Command-line interface output"""

from image_creator.output import Output

import sys
import time
from colors import red, green, yellow
from progress.bar import Bar


def write(msg, new_line, decorate, stream):
    """Print a message"""
    nl = "\n" if new_line else ' '
    stream.write(decorate(msg) + nl)


class SilentOutput(Output):
    """Silent Output class. Only Errors are printed"""

    def __init__(self, **kwargs):
        """Initialize a SilentOutput instance"""
        self.colored = kwargs['colored'] if 'colored' in kwargs else True
        self.stdout = kwargs['stdout'] if 'stdout' in kwargs else sys.stdout
        self.stderr = kwargs['stderr'] if 'stderr' in kwargs else sys.stderr

        if 'timestamp' in kwargs and kwargs['timestamp']:
            self.timestamp = lambda x: "[%ds] %s" % \
                ((time.time() - self.start), x)
        else:
            self.timestamp = lambda x: x

        self.yellow = yellow if self.colored else lambda x: x
        self.red = red if self.colored else lambda x: x
        self.green = green if self.colored else lambda x: x

        # Are we in the beginning of a line?
        self.linestart = True

    def result(self, msg):
        """Print a result"""
        write(msg, True, lambda x: x, self.stdout)
        self.linestart = True

    def error(self, msg):
        """Print an error"""
        decorator = (lambda x: self.red(self.timestamp(x))) \
            if self.linestart else self.red
        write("Error: %s" % msg, True, decorator, self.stderr)
        self.linestart = True


class SimpleOutput(SilentOutput):
    """Print messages but not progress bars. Progress bars are treated as
    output messages. The user gets informed when the action begins and when it
    ends, but no progress is shown in between."""

    def warn(self, msg):
        """Print a warning"""
        decorator = (lambda x: self.yellow(self.timestamp(x))) \
            if self.linestart else self.yellow
        write("Warning: %s" % msg, True, decorator, self.stderr)
        self.linestart = True

    def success(self, msg):
        """Print msg after an action is completed"""
        decorator = (lambda x: self.green(self.timestamp(x))) \
            if self.linestart else self.green
        write(msg, True, decorator, self.stderr)
        self.linestart = True

    def info(self, msg='', new_line=True):
        """Print msg as normal program output"""
        decorator = (lambda x: self.timestamp(x)) \
            if self.linestart else lambda x: x
        write(msg, new_line, decorator, self.stderr)
        self.linestart = new_line

    def clear(self):
        """Clear the screen"""
        if self.stderr.isatty():
            self.stderr.write('\033[H\033[2J')
        else:
            self.info()


class OutputWthProgress(SimpleOutput):
    """Output class with progress."""
    class _Progress(Bar):
        MESSAGE_LENGTH = 30

        template = {
            'default': '%(index)d/%(max)d',
            'percent': '%(percent)d%%',
            'b': '%(index)d/%(max)d B',
            'kb': '%(index)d/%(max)d KB',
            'mb': '%(index)d/%(max)d MB'
        }

        def __init__(self, size, title, bar_type='default'):
            """Create a Progress bar"""
            self.hide_cursor = False
            super(OutputWthProgress._Progress, self).__init__()
            self.title = title
            self.fill = '#'
            self.bar_prefix = ' ['
            self.bar_suffix = '] '
            self.suffix = self.template[bar_type]
            self.max = size
            self.title = self.parent.timestamp(title)
            self.message = ("%s:" % self.title).ljust(self.MESSAGE_LENGTH)

            # print empty progress bar
            self.start()

        def success(self, result):
            """Print result after progress has finished"""
            self.parent.info("\r%s ...\033[K" % self.title, False)
            self.parent.success(result)

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2015 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module hosts OS-specific code for Ubuntu Linux."""

import re

from image_creator.os_type.linux import Linux


class Ubuntu(Linux):
    """OS class for Ubuntu Linux variants"""

    def _do_collect_metadata(self):
        """Collect metadata about the OS"""

        super(Ubuntu, self)._do_collect_metadata()

        regexp = re.compile('^(k|l|x)ubuntu-desktop$')
        variant = ""
        for app in self.image.g.inspect_list_applications(self.root):
            match = regexp.match(app['app_name'])
            if match:
                variant = match.group(1) + 'ubuntu'
                break

        if variant:
            self.meta['OS'] = variant
            descr = self.meta['DESCRIPTION'].replace('Ubuntu',
                                                     variant.capitalize())
            self.meta['DESCRIPTION'] = descr

        # Check if this is a bitnami image
        if self.image.g.is_dir('/opt/bitnami'):
            self.meta['OS'] = 'bitnami'
            readme = '/opt/bitnami/README.txt'
            if self.image.g.is_file(readme):
                content = self.image.g.cat(readme).splitlines()
                if len(content):
                    self.meta['OSVERSION'] = self.meta['DESCRIPTION']
                    self.meta['DESCRIPTION'] = content[0].strip()


# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2015 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module is the entrance point for the dialog-based version of the
snf-image-creator program. The main function will create a dialog where the
user is asked if he wants to use the program in expert or wizard mode.
"""

from __future__ import unicode_literals

import dialog
import sys
import os
import signal
import argparse
import types
import termios
import traceback
import tempfile

from image_creator import __version__ as version
from image_creator.util import FatalError
from image_creator.output.cli import SimpleOutput
from image_creator.output.dialog import GaugeOutput
from image_creator.output.composite import CompositeOutput
from image_creator.output.syslog import SyslogOutput
from image_creator.disk import Disk
from image_creator.dialog_wizard import start_wizard
from image_creator.dialog_menu import main_menu
from image_creator.dialog_util import WIDTH, confirm_exit, Reset, \
    update_background_title, select_file

PROGNAME = os.path.basename(sys.argv[0])


def create_image(d, media, out, tmp, snapshot):
    """Create an image out of `media'"""
    d.setBackgroundTitle('snf-image-creator')

    gauge = GaugeOutput(d, "Initialization", "Initializing...")
    out.append(gauge)
    disk = Disk(media, out, tmp)

    def signal_handler(signum, frame):
        gauge.cleanup()
        disk.cleanup()

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    try:

        device = disk.file if not snapshot else disk.snapshot()

        image = disk.get_image(device)

        gauge.cleanup()
        out.remove(gauge)

        # Make sure the signal handler does not call gauge.cleanup again
        def dummy(self):
            pass
        gauge.cleanup = type(GaugeOutput.cleanup)(dummy, gauge, GaugeOutput)

        session = {"dialog": d,
                   "disk": disk,
                   "image": image}

        if image.is_unsupported():

            session['excluded_tasks'] = [-1]
            session['task_metadata'] = ["EXCLUDE_ALL_TASKS"]

            msg = "The system on the input media is not supported." \
                "\n\nReason: %s\n\n" \
                "We highly recommend not to create an image out of this, " \
                "since the image won't be cleaned up and you will not be " \
                "able to configure it during the deployment. Press <YES> if " \
                "you still want to continue with the image creation process." \
                % image._unsupported

            if d.yesno(msg, width=WIDTH, defaultno=1, height=12) == d.OK:
                main_menu(session)

            d.infobox("Thank you for using snf-image-creator. Bye", width=53)
            return 0

        msg = "snf-image-creator detected a %s system on the input media. " \
              "Would you like to run a wizard to assist you through the " \
              "image creation process?\n\nChoose <Wizard> to run the wizard," \
              " <Expert> to run snf-image-creator in expert mode or press " \
              "ESC to quit the program." \
              % (image.ostype.capitalize() if image.ostype == image.distro or
                 image.distro == "unknown" else "%s (%s)" %
                 (image.ostype.capitalize(), image.distro.capitalize()))

        update_background_title(session)

        while True:
            code = d.yesno(msg, width=WIDTH, height=12, yes_label="Wizard",
                           no_label="Expert")
            if code == d.OK:
                if start_wizard(session):
                    break
            elif code == d.CANCEL:
                main_menu(session)
                break

            if confirm_exit(d):
                break

        d.infobox("Thank you for using snf-image-creator. Bye", width=53)
    finally:
        disk.cleanup()

    return 0


def _dialog_form(self, text, elements, height=20, width=60, form_height=15,
                 **kwargs):
    """Display a form box.
       Each element of *elements* must itself be a sequence
       :samp:`({label}, {yl}, {xl}, {item}, {yi}, {xi}, {field_length},
       {input_length})` containing the various parameters concerning a
       given field and the associated label.
       *label* is a string that will be displayed at row *yl*, column
       *xl*. *item* is a string giving the initial value for the field,
       which will be displayed at row *yi*, column *xi* (row and column
       numbers starting from 1).
       *field_length* and *input_length* are integers that respectively
       specify the number of characters used for displaying the field
       and the maximum number of characters that can be entered for
       this field. These two integers also determine whether the
       contents of the field can be modified, as follows:
         - if *field_length* is zero, the field cannot be altered and
           its contents determines the displayed length;
         - if *field_length* is negative, the field cannot be altered
           and the opposite of *field_length* gives the displayed
           length;
         - if *input_length* is zero, it is set to *field_length*.
    """

    cmd = ["--form", text, str(height), str(width), str(form_height)]

    for element in elements:
        label, yl, xl, item, yi, xi, field_len, input_len = element[:8]

        cmd.extend((label, unicode(yl), unicode(xl), item, unicode(yi),
                    unicode(xi), unicode(field_len), unicode(input_len)))

    code, output = self._perform(*(cmd,), **kwargs)

    if not output:
        return (code, [])

    return (code, output.splitlines())


def dialog_main(media, **kwargs):
    """Main function for the dialog-based version of the program"""

    tmpdir = kwargs['tmpdir'] if 'tmpdir' in kwargs else None
    snapshot = kwargs['snapshot'] if 'snapshot' in kwargs else True
    logfile = kwargs['logfile'] if 'logfile' in kwargs else None
    syslog = kwargs['syslog'] if 'syslog' in kwargs else False

    # In openSUSE dialog is buggy under xterm
    if os.environ['TERM'] == 'xterm':
        os.environ['TERM'] = 'linux'

    d = dialog.Dialog(dialog="dialog")

    # Add extra button in dialog library if missing
    if 'extra_button' not in dialog._common_args_syntax:
        dialog._common_args_syntax["extra_button"] = \
            lambda enable: dialog._simple_option("--extra-button", enable)
    if 'extra_label' not in dialog._common_args_syntax:
        dialog._common_args_syntax["extra_label"] = \
            lambda string: ("--extra-label", string)

    # Allow yes-no label overwriting if missing
    if 'yes_label' not in dialog._common_args_syntax:
        dialog._common_args_syntax["yes_label"] = \
            lambda string: ("--yes-label", string)
    if 'no_label' not in dialog._common_args_syntax:
        dialog._common_args_syntax["no_label"] = \
            lambda string: ("--no-label", string)

    # Add exit label overwriting if missing
    if 'exit_label' not in dialog._common_args_syntax:
        dialog._common_args_syntax["exit_label"] = \
            lambda string: ("--exit-label", string)

    # Monkey-patch pythondialog to include support for form dialog boxes
    if not hasattr(d, 'form'):
        d.form = types.MethodType(_dialog_form, d)

    # Add sort dialog constants if missing
    if not hasattr(d, 'OK'):
        d.OK = d.DIALOG_OK

    if not hasattr(d, 'CANCEL'):
        d.CANCEL = d.DIALOG_CANCEL

    if not hasattr(d, 'ESC'):
        d.ESC = d.DIALOG_ESC

    if not hasattr(d, 'EXTRA'):
        d.EXTRA = d.DIALOG_EXTRA

    if not hasattr(d, 'HELP'):
        d.HELP = d.DIALOG_HELP

    d.setBackgroundTitle('snf-image-creator')

    # Pick input media
    while True:
        media = select_file(d, init=media, ftype="br", bundle_host=True,
                            title="Please select an input media.")
        if media is None:
            if confirm_exit(
                    d, "You canceled the media selection dialog box."):
                return 0
            continue
        break

    tmplog = None if logfile else tempfile.NamedTemporaryFile(prefix='fatal-',
                                                              delete=False)

    logs = []
    try:
        stream = logfile if logfile else tmplog
        logs.append(SimpleOutput(colored=False, stderr=stream, stdout=stream,
                                 timestamp=True))
        if syslog:
            logs.append(SyslogOutput())

        while 1:
            try:
                out = CompositeOutput(logs)
                out.info("Starting %s v%s ..." % (PROGNAME, version))
                ret = create_image(d, media, out, tmpdir, snapshot)
                break
            except Reset:
                for log in logs:
                    log.info("Resetting everything ...")
    except FatalError as error:
        for log in logs:
            log.error(str(error))
        msg = 'A fatal error occured. See %s for a full log.' % log.stderr.name
        d.infobox(msg, width=WIDTH, title="Fatal Error")
        return 1
    else:
        if tmplog:
            os.unlink(tmplog.name)
    finally:
        if tmplog:
            tmplog.close()

    return ret


def main():
    """Entrance Point"""
    if os.geteuid() != 0:
        sys.stderr.write("Error: You must run %s as root\n" % PROGNAME)
        sys.exit(2)

    description = "Dialog-based tool for creating OS images"
    parser = argparse.ArgumentParser(version=version, description=description)
    parser.add_argument("-l", "--logfile", dest="logfile", metavar="FILE",
                        default=None, help="log all messages to FILE")
    parser.add_argument("--no-snapshot", dest="snapshot", default=True,
                        help="don't snapshot the input media. (THIS IS "
                        "DANGEROUS AS IT WILL ALTER THE ORIGINAL MEDIA!!!)",
                        action="store_false")
    parser.add_argument("--syslog", dest="syslog", default=False,
                        help="log to syslog", action="store_true")
    parser.add_argument("--tmpdir", dest="tmp", default=None, metavar="DIR",
                        help="create large temporary image files under DIR")
    parser.add_argument("source", metavar="SOURCE", default=None, nargs='?',
                        help="Image file, block device or /")

    opts = parser.parse_args()

    if opts.tmp is not None and not os.path.isdir(opts.tmp):
        parser.error("Directory: `%s' specified with --tmpdir is not valid"
                     % opts.tmp)

    try:
        logfile = open(opts.logfile, 'w') if opts.logfile is not None else None
    except IOError as error:
        parser.error("Unable to open logfile `%s' for writing. Reason: %s" %
                     (opts.logfile, error.strerror))

    try:
        # Save the terminal attributes
        attr = termios.tcgetattr(sys.stdin.fileno())
        try:
            ret = dialog_main(opts.source, logfile=logfile, tmpdir=opts.tmp,
                              snapshot=opts.snapshot, syslog=opts.syslog)
        finally:
            # Restore the terminal attributes. If an error occurs make sure
            # that the terminal turns back to normal.
            termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN, attr)
    except:
        # Clear the screen
        sys.stdout.write('\033[2J')  # Erase Screen
        sys.stdout.write('\033[H')  # Cursor Home
        sys.stdout.flush()

        exception = traceback.format_exc()
        sys.stderr.write(exception)
        if logfile is not None:
            logfile.write(exception)

        sys.exit(3)
    finally:
        if logfile is not None:
            logfile.close()

    sys.exit(ret)

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2015 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides the Syslog output class"""

from __future__ import absolute_import
from image_creator.output import Output
import syslog


class SyslogOutput(Output):
    """This class implements the Output interface for the system logger"""
    def __init__(self, ident=None, logoption=None, facility=None):
        """Initialize a Syslog instance"""

        self.ident = ident if ident is not None else 'snf-image-creator'
        self.logoption = logoption if logoption is not None else syslog.LOG_PID
        self.facility = facility if facility is not None else syslog.LOG_USER
        self.line = []

        syslog.openlog(self.ident, self.logoption, self.facility)

    def error(self, msg):
        """Print an error to syslog"""

        if len(self.line):
            syslog.syslog(syslog.LOG_INFO, " ".join(self.line))
            self.line = []

        syslog.syslog(syslog.LOG_ERR, "[ERROR] %s" % msg)

    def warn(self, msg):
        """Print a warning"""

        if len(self.line):
            syslog.syslog(syslog.LOG_INFO, "[INFO] %s " % " ".join(self.line))
            self.line = []

        syslog.syslog(syslog.LOG_WARNING, "[WARNING] %s" % msg)

    def success(self, msg):
        """Print msg after an action is completed"""

        self.line.append(msg)
        syslog.syslog(syslog.LOG_INFO, "[INFO] %s" % " ".join(self.line))
        self.line = []

    def info(self, msg='', new_line=True):
        """Print normal program output"""

        if len(msg) == 0:
            return

        self.line.append(msg)
        if new_line:
            syslog.syslog(syslog.LOG_INFO, "[INFO] %s" % " ".join(self.line))
            self.line = []

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module hosts OS-specific code for FreeBSD."""

from image_creator.os_type.bsd import Bsd

import re


class Freebsd(Bsd):
    """OS class for FreeBSD Unix-like operating system"""

    def _check_enabled_sshd(self):
        """Check if the sshd is enabled at boot"""

        sshd_enabled = False
        sshd_service = re.compile(r'^sshd_enable=.+$')

        # Freebsd has a checkyesno() functions that tests the service variable
        # against all those different values in a case insensitive manner!!!
        sshd_yes = re.compile(r"^sshd_enable=(['\"]?)(YES|TRUE|ON|1)\1$",
                              re.IGNORECASE)
        for rc_conf in ('/etc/rc.conf', '/etc/rc.conf.local'):
            if not self.image.g.is_file(rc_conf):
                continue

            for line in self.image.g.cat(rc_conf).splitlines():
                line = line.split('#')[0].strip()
                # Be paranoid. Don't stop examining lines after a match. This
                # is a shell variable and can be overwritten many times. Only
                # the last match counts.
                if sshd_service.match(line):
                    sshd_enabled = sshd_yes.match(line) is not None

        return sshd_enabled

    def _do_inspect(self):
        """Run various diagnostics to check if media is supported"""

        self.out.info('Checking partition table type...', False)
        ptype = self.image.g.part_get_parttype(self.image.guestfs_device)
        if ptype != 'gpt':
            self.out.warn("partition table type is: `%s'" % ptype)
            self.image.set_unsupported(
                'On FreeBSD only GUID partition tables are supported')
        else:
            self.out.success(ptype)

    def _do_mount(self, readonly):
        """Mount partitions in the correct order"""

        critical_mpoints = ('/', '/etc', '/root', '/home', '/var')

        # Older libguestfs versions can't handle correct FreeBSD partitions on
        # a GUID Partition Table. We have to do the translation to Linux device
        # names ourselves
        guid_device = re.compile(r'^/dev/((?:ada)|(?:vtbd))(\d+)p(\d+)$')

        mopts = "ufstype=ufs2,%s" % ('ro' if readonly else 'rw')
        for mp, dev in self._mountpoints():
            match = guid_device.match(dev)
            if match:
                group2 = int(match.group(2))
                group3 = int(match.group(3))
                dev = '/dev/sd%c%d' % (chr(ord('a') + group2), group3)
            try:
                self.image.g.mount_vfs(mopts, 'ufs', dev, mp)
            except RuntimeError as msg:
                if mp in critical_mpoints:
                    self._mount_error = str(msg)
                    return False
                else:
                    self._mount_warnings.append('%s (ignored)' % msg)

        return True

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module hosting code for determining the installed boot loader on an image.
The bootloader signatures are taken from the "Boot Info Script" project:
    https://github.com/arvidjaar/bootinfoscript
"""

# Master Boot Record Bootloaders
MBR_LDR = [
    # GRUB Legacy
    "grub1",            # 0
    # GRUB 2
    "grub2",
    # LILO (LInux LOader)
    "lilo",
    # SYSLINUX
    "syslinux",
    # Windows
    "Windows",
    # FreeBSD's boot0
    "freebsd",          # 5
    # NetBSD's Stage 0/SUSE generic MBR
    "netbsd",
    # OpenBSD's MBR
    "openbsd",
    # ReactOS
    "reactos",
    # FreeDOS
    "freedos",
    # MS-DOS
    "msdos",            # 10
    # Solaris
    "solaris",
    # ThinkPad MBR
    "thinkpad",
    # HP/Gateway
    "hp",
    # Plop Boot Manager
    "plop",
    # TrueCrypt boot loader
    "truecrypt",        # 15
    # Paragon Partition Manager
    "paragon",
    # Testdisk MBR
    "testdist",
    # GAG Graphical Boot Manager
    "gag",
    # BootIt Boot Manager
    "bootit",
    # DiskCryptor
    "diskcryptor",  # 20
    # xOSL (Extended Operating System Loader)
    "xosl",
    # Fbinst
    "fbinst",
    # Grub4Dos
    "grub4dos",
    # WEE boot manager
    "wee",
    # mbldr (Master Boot LoaDeR)
    "mbldr",   # 25
    # Libparted generic boot code
    "libparted",
    # ISOhybrid
    "isohybrid",
    # Acer PQservice MBR
    "pqservice",
]

# 2-bytes MBR signature
MBR_SIG2 = {
    "\x3b\x48": 0,
    "\xeb\x4c": 1,
    "\xeb\x63": 1,
    "\xeb\x04": 11,
    "\x0e\xbe": 12,
    "\x33\xed": 27,
    "\x33\xff": 13,
    "\xb8\x00": 14,
    "\xea\x1e": 15,
    "\xeb\x04":        11,
    "\xeb\x31": 16,
    "\xfa\x33": 10,
    "\xfa\xeb": 2,
    "\xfa\xfc": 8,
    "\xfc\x31": 17,
    "\xfc\x33": 18,
    "\xfc\xeb": 19,
}

# 3-bytes MBR signature
MBR_SIG3 = {
    "\x33\xc0\x8e": 4,
    "\x33\xc0\x90": 20,
    "\x33\xc0\xfa": 3,
    "\xea\x05\x00": 7,
    "\xea\x05\x01": 21,
    "\xeb\x5e\x00": 22,
    "\xeb\x5e\x80": 23,
    "\xeb\x5e\x90": 24,
    "\xfa\x31\xc0": 3,
    "\xfa\x31\xc9": 25,
    "\xfa\x31\xed": 27,
}

# 4-bytes MBR signature
MBR_SIG4 = {
    "\xfa\xb8\x00\x00": 9,
    "\xfa\xb8\x00\x10": 26,
}

# 8-bytes MBR signature
MBR_SIG8 = {
    "\x31\xc0\x8e\xd0\xbc\x00\x7c\x8e": 6,
    "\x31\xc0\x8e\xd0\xbc\x00\x7c\xfb": 28,
}


def mbr_bootinfo(mbr):
    """Inspect a Master Boot Record and return the installed bootloader"""
    if mbr[:2] == '\x00\x00':
        return "none"

    try:
        ret = MBR_SIG2[mbr[:2]]
    except KeyError:
        try:
            ret = MBR_SIG3[mbr[:3]]
        except KeyError:
            try:
                ret = MBR_SIG4[mbr[:4]]
            except KeyError:
                try:
                    ret = MBR_SIG8[mbr[:8]]
                except KeyError:
                    return "unknown"

    return MBR_LDR[ret]


# Volume Boot Record Bootloaders
VBR_LDR = [
    # GRUB Legacy
    "grub1",        # 0
    # GRUB 2
    "grub2",
    # LILO (LInux LOader)
    "lilo"
    # SYSLINUX
    "syslinux",
    # Windows
    "windows",
    # FreeBSD's boot1 (First sector of FreeBSD slice)
    "freebsd",      # 5
    # NetBSD's PBR
    "netbsd",
    # OpenBSD
    "openbsd",
    # ReactOS
    "reactos",
    # FreeDOS
    "freedos",
    # MS-DOS
    "msdos",        # 10
    # HP Recovery
    "hp",
    # Dell Utility/Recovery
    "dell",
    # ISOhybrid
    "isohybrid",
    # Grub4Dos
    "grub4dos",
    # BootIt Boot Manager
    "bootit",       # 15
    # Acronis Secure Zone
    "acronis_sz",
]

# 2-byte VBR "no bootloader" signature
VBR_SIG2_NONE = (
    "\x00\x00",
    "\x55\xcd",
    "\x69\x6e",
    "\x6f\x6e",
)

# 2-byte VBR signature
VBR_SIG2 = {
    "\x00\x20": 1,
    "\x00\x69": 13,
    "\x01\x0f": 11,
    "\x02\x11": 12,
    "\x04\x88": 1,
    "\x06\x89": 3,
    "\x07\x34": 10,
    "\x07\x45": 4,
    "\x08\x9e": 10,
    "\x08\xcd": 4,
    "\x08\xc7": 7,
    "\x0b\x60": 12,
    "\x0b\xd0": 4,
    "\x0e\x00": 12,
    "\x0f\xb6": 13,
    "\x2a\x00": 8,
    "\x2d\x5e": 10,
    "\x31\xc0": 3,
    "\x3c\x9a": 6,
    "\x40\x7c": 3,
    "\x42\x16": 14,
    "\x44\x45": 12,
    "\x48\xb4": 1,
    "\x52\x72": 0,
    "\x56\x26": 14,
    "\x63\x8b": 9,
    "\x66\x16": 4,
    "\x69\x74": 15,
    "\x6a\x02": 5,
    "\x6f\x65": 15,
    "\x74\x05": 4,
    "\x7c\x3c": 1,
    "\x7c\xc6": 4,
    "\x7e\x1e": 14,
    "\x80\x53": 2,
    "\x8a\x56": 16,
    "\x83\xe1": 13,
    "\x8e\xc0": 4,
    "\x8e\xd0": 12,
    "\xaa\x75": 0,
    "\xb1\x06": 3,
    "\xb6\x00": 12,
    "\xb6\xc6": 13,
    "\xb6\xd1": 4,
    "\xe8\x79": 13,
    "\xe9\xd8": 4,
    "\xf6\xc1": 4,
    "\xfa\x33": 4,
}

# 4-byte VBR signature
VBR_SIG4 = {
    "\x55\xaa\x75\x0a": 14,
    "\x55\xaa\x75\x06": 4,
    "\x78\x15\xb1\x06": 3,
}


def vbr_bootinfo(vbr):
    """Inspect a Volume Boot Record and return the installed bootloader"""
    if vbr[0x80:0x82] in VBR_SIG2_NONE:
        return "none"

    try:
        ret = VBR_SIG2[vbr[0x80:0x82]]
    except KeyError:
        try:
            ret = VBR_SIG4[vbr[0x80:0x84]]
        except KeyError:
            return "unknown"

    return MBR_LDR[ret]

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2015 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module hosting the Image class."""

from image_creator.util import FatalError, QemuNBD, get_command
from image_creator.gpt import GPTPartitionTable
from image_creator.os_type import os_cls

import os
import re
import hashlib
from sendfile import sendfile
import threading

# Make sure libguestfs runs qemu directly to launch an appliance.
os.environ['LIBGUESTFS_BACKEND'] = 'direct'
import guestfs  # noqa


class Image(object):
    """The instances of this class can create images out of block devices."""

    def __init__(self, device, output, **kwargs):
        """Create a new Image instance"""

        self.device = device
        self.out = output
        self.format = kwargs['format'] if 'format' in kwargs else 'raw'

        self.meta = kwargs['meta'] if 'meta' in kwargs else {}
        self.sysprep_params = \
            kwargs['sysprep_params'] if 'sysprep_params' in kwargs else {}

        self.progress_bar = None
        self.guestfs_device = None
        self.size = 0

        self.g = guestfs.GuestFS()
        self.guestfs_enabled = False
        self.guestfs_version = self.g.version()

        # This is needed if the image format is not raw
        self.nbd = QemuNBD(device)

        if self.nbd.qemu_nbd is None and self.format != 'raw':
            raise FatalError("qemu-nbd command is missing, only raw input "
                             "media are supported")

        # Check If MOUNT LOCAL is supported for this guestfs build
        self.mount_local_support = hasattr(self.g, "mount_local")
        if self.mount_local_support:
            self._mount_thread = None

    def check_guestfs_version(self, major, minor, release):
        """Checks if the version of the used libguestfs is smaller, equal or
        greater than the one specified by the major, minor and release triplet

        Returns:
            < 0 if the installed version is smaller than the specified one
            = 0 if they are equal
            > 0 if the installed one is greater than the specified one
        """

        for (a, b) in (self.guestfs_version['major'], major), \
                (self.guestfs_version['minor'], minor), \
                (self.guestfs_version['release'], release):
            if a != b:
                return a - b

        return 0

    def enable(self):
        """Enable a newly created Image instance"""

        self.enable_guestfs()

        self.out.info('Inspecting Operating System ...', False)
        roots = self.g.inspect_os()

        if len(roots) == 0 or len(roots) > 1:
            self.root = None
            self.ostype = "unsupported"
            self.distro = "unsupported"
            self.guestfs_device = '/dev/sda'
            self.size = self.g.blockdev_getsize64(self.guestfs_device)

            if len(roots) > 1:
                reason = "Multiple operating systems found on the media."
            else:
                reason = "Unable to detect any operating system on the media."

            self.set_unsupported(reason)
            return

        self.root = roots[0]
        self.meta['PARTITION_TABLE'] = self.g.part_get_parttype('/dev/sda')
        self.guestfs_device = '/dev/sda'  # self.g.part_to_dev(self.root)
        self.size = self.g.blockdev_getsize64(self.guestfs_device)

        self.ostype = self.g.inspect_get_type(self.root)
        self.distro = self.g.inspect_get_distro(self.root)
        self.out.success(
            'found a(n) %s system' %
            self.ostype if self.distro == "unknown" else self.distro)

        # Inspect the OS
        self.os.inspect()

    def set_unsupported(self, reason):
        """Flag this image as unsupported"""

        self._unsupported = reason
        self.meta['UNSUPPORTED'] = reason
        self.out.warn('Media is not supported. Reason: %s' % reason)

    def is_unsupported(self):
        """Returns if this image is unsupported"""
        return hasattr(self, '_unsupported')

    def enable_guestfs(self):
        """Enable the guestfs handler"""

        if self.guestfs_enabled:
            self.out.warn("Guestfs is already enabled")
            return

        # Before version 1.18.4 the behavior of kill_subprocess was different
        # and you need to reset the guestfs handler to relaunch a previously
        # shut down QEMU backend
        if self.check_guestfs_version(1, 18, 4) < 0:
            self.g = guestfs.GuestFS()

        self.g.add_drive_opts(self.device, readonly=0)

        # Before version 1.17.14 the recovery process, which is a fork of the
        # original process that called libguestfs, did not close its inherited
        # file descriptors. This can cause problems especially if the parent
        # process has opened pipes. Since the recovery process is an optional
        # feature of libguestfs, it's better to disable it.
        if self.check_guestfs_version(1, 17, 14) >= 0:
            self.out.info("Enabling recovery process ...", False)
            self.g.set_recovery_proc(1)
            self.out.success('done')
        else:
            self.g.set_recovery_proc(0)

        # self.g.set_trace(1)
        # self.g.set_verbose(1)

        self.out.info('Launching helper VM (may take a while) ...', False)
        # self.progressbar = self.out.Progress(100, "Launching helper VM",
        #                                     "percent")
        # eh = self.g.set_event_callback(self.progress_callback,
        #                               guestfs.EVENT_PROGRESS)
        try:
            self.g.launch()
        except RuntimeError as e:
            raise FatalError(
                "Launching libguestfs's helper VM failed!\nReason: %s.\n\n"
                "Please run `libguestfs-test-tool' for more info." % str(e))

        self.guestfs_enabled = True
        # self.g.delete_event_callback(eh)
        # self.progressbar.success('done')
        # self.progressbar = None

        if self.check_guestfs_version(1, 18, 4) < 0:
            self.g.inspect_os()  # some calls need this

        self.out.success('done')

    def disable_guestfs(self):
        """Disable the guestfs handler"""

        if not self.guestfs_enabled:
            self.out.warn("Guestfs is already disabled")
            return

        self.out.info("Shutting down helper VM ...", False)
        self.g.sync()
        # guestfs_shutdown which is the preferred way to shutdown the backend
        # process was introduced in version 1.19.16
        if self.check_guestfs_version(1, 19, 16) >= 0:
            self.g.shutdown()
        else:
            self.g.kill_subprocess()

        # We will reset the guestfs handler if needed
        if self.check_guestfs_version(1, 18, 4) < 0:
            self.g.close()

        self.guestfs_enabled = False
        self.out.success('done')

    @property
    def os(self):
        """Return an OS class instance for this image"""
        if hasattr(self, "_os"):
            return self._os

        if not self.guestfs_enabled:
            self.enable()

        cls = os_cls(self.distro, self.ostype)
        self._os = cls(self, sysprep_params=self.sysprep_params)

        self._os.collect_metadata()

        return self._os

    def raw_device(self, readonly=True):
        """Returns a context manager that exports the raw image device. If
        readonly is true, the block device that is returned is read only.
        """

        if self.guestfs_enabled:
            self.g.umount_all()
            self.g.sync()
            self.g.drop_caches(3)  # drop everything

        # Self gets overwritten
        img = self

        class RawImage(object):
            """The RawImage context manager"""
            def __enter__(self):
                return img.device if img.format == 'raw' else \
                    img.nbd.connect(readonly)

            def __exit__(self, exc_type, exc_value, traceback):
                if img.format != 'raw':
                    img.nbd.disconnect()

        return RawImage()

    def destroy(self):
        """Destroy this Image instance."""

        # In new guestfs versions, there is a handy shutdown method for this
        try:
            if self.guestfs_enabled:
                self.g.umount_all()
                self.g.sync()
        finally:
            # Close the guestfs handler if open
            self.g.close()

#    def progress_callback(self, ev, eh, buf, array):
#        position = array[2]
#        total = array[3]
#
#        self.progressbar.goto((position * 100) // total)

    def _last_partition(self):
        """Return the last partition of the image disk"""
        if self.meta['PARTITION_TABLE'] not in 'msdos' 'gpt':
            msg = "Unsupported partition table: %s. Only msdos and gpt " \
                "partition tables are supported" % self.meta['PARTITION_TABLE']
            raise FatalError(msg)

        def is_extended(partition):
            return self.g.part_get_mbr_id(
                self.guestfs_device, partition['part_num']) in (0x5, 0xf)

        def is_logical(partition):
            return self.meta['PARTITION_TABLE'] == 'msdos' and \
                partition['part_num'] > 4

        partitions = self.g.part_list(self.guestfs_device)
        last_partition = partitions[-1]

        if is_logical(last_partition):
            # The disk contains extended and logical partitions....
            extended = [p for p in partitions if is_extended(p)][0]
            last_primary = [p for p in partitions if p['part_num'] <= 4][-1]

            # check if extended is the last primary partition
            if last_primary['part_num'] > extended['part_num']:
                last_partition = last_primary

        return last_partition

    def shrink(self, silent=False):
        """Shrink the image.

        This is accomplished by shrinking the last file system of the
        image and then updating the partition table. The shrinked device is
        returned.

        ATTENTION: make sure umount is called before shrink
        """
        def get_fstype(partition):
            """Get file system type"""
            device = "%s%d" % (self.guestfs_device, partition['part_num'])
            return self.g.vfs_type(device)

        def is_logical(partition):
            """Returns True if the partition is a logical partition"""
            return self.meta['PARTITION_TABLE'] == 'msdos' and \
                partition['part_num'] > 4

        def is_extended(partition):
            """Returns True if the partition is an extended partition"""
            if self.meta['PARTITION_TABLE'] == 'msdos':
                mbr_id = self.g.part_get_mbr_id(self.guestfs_device,
                                                partition['part_num'])
                return mbr_id in (0x5, 0xf)
            return False

        def part_add(ptype, start, stop):
            """Add partition"""
            self.g.part_add(self.guestfs_device, ptype, start, stop)

        def part_del(partnum):
            """Delete a partition"""
            self.g.part_del(self.guestfs_device, partnum)

        def part_get_id(partnum):
            """Returns the MBR id of the partition"""
            return self.g.part_get_mbr_id(self.guestfs_device, partnum)

        def part_set_id(partnum, id):
            """Sets the MBR id of the partition"""
            self.g.part_set_mbr_id(self.guestfs_device, partnum, id)

        def part_get_bootable(partnum):
            """Returns the bootable flag of the partition"""
            return self.g.part_get_bootable(self.guestfs_device, partnum)

        def part_set_bootable(partnum, bootable):
            """Sets the bootable flag for a partition"""
            self.g.part_set_bootable(self.guestfs_device, partnum, bootable)

        MB = 2 ** 20

        if self.is_unsupported():
            if not silent:
                self.out.warn("Shrinking is disabled for unsupported images")
            return None

        sector_size = self.g.blockdev_getss(self.guestfs_device)

        last_part = None
        fstype = None
        while True:
            last_part = self._last_partition()
            fstype = get_fstype(last_part)

            if fstype == 'swap':
                self.meta['SWAP'] = "%d:%s" % \
                    (last_part['part_num'],
                     (last_part['part_size'] + MB - 1) // MB)
                part_del(last_part['part_num'])
                continue
            elif is_extended(last_part):
                part_del(last_part['part_num'])
                continue

            # Most disk manipulation programs leave 2048 sectors after the last
            # partition
            new_size = last_part['part_end'] + 1 + 2048 * sector_size
            self.size = min(self.size, new_size)
            break

        if not re.match("ext[234]", fstype):
            if not silent:
                self.out.warn(
                    "Don't know how to shrink %s partitions." % fstype)
            return None

        part_dev = "%s%d" % (self.guestfs_device, last_part['part_num'])

        try:
            if self.check_guestfs_version(1, 15, 17) >= 0:
                self.g.e2fsck(part_dev, forceall=1)
            else:
                self.g.e2fsck_f(part_dev)
        except RuntimeError as e:
            # There is a bug in some versions of libguestfs and a RuntimeError
            # is thrown although the command has successfully corrected the
            # found file system errors.
            if e.message.find('***** FILE SYSTEM WAS MODIFIED *****') == -1:
                raise

        self.g.resize2fs_M(part_dev)

        out = self.g.tune2fs_l(part_dev)
        block_size = int(filter(lambda x: x[0] == 'Block size', out)[0][1])
        block_cnt = int(filter(lambda x: x[0] == 'Block count', out)[0][1])

        start = last_part['part_start'] / sector_size
        end = start + (block_size * block_cnt) / sector_size - 1

        if is_logical(last_part):
            partitions = self.g.part_list(self.guestfs_device)

            logical = []  # logical partitions
            for partition in partitions:
                if partition['part_num'] < 4:
                    continue
                logical.append({
                    'num': partition['part_num'],
                    'start': partition['part_start'] / sector_size,
                    'end': partition['part_end'] / sector_size,
                    'id': part_get_id(partition['part_num']),
                    'bootable': part_get_bootable(partition['part_num'])
                })

            logical[-1]['end'] = end  # new end after resize

            # Recreate the extended partition
            extended = filter(is_extended, partitions)[0]
            part_del(extended['part_num'])
            part_add('e', extended['part_start'] / sector_size, end)

            # Create all the logical partitions back
            for l in logical:
                part_add('l', l['start'], l['end'])
                part_set_id(l['num'], l['id'])
                part_set_bootable(l['num'], l['bootable'])
        else:
            # Recreate the last partition
            if self.meta['PARTITION_TABLE'] == 'msdos':
                last_part['id'] = part_get_id(last_part['part_num'])

            last_part['bootable'] = part_get_bootable(last_part['part_num'])
            part_del(last_part['part_num'])
            part_add('p', start, end)
            part_set_bootable(last_part['part_num'], last_part['bootable'])

            if self.meta['PARTITION_TABLE'] == 'msdos':
                part_set_id(last_part['part_num'], last_part['id'])

        new_size = (end + 1) * sector_size

        assert new_size <= self.size

        if self.meta['PARTITION_TABLE'] == 'gpt':
            with self.raw_device(readonly=False) as raw:
                ptable = GPTPartitionTable(raw)
                self.size = ptable.shrink(new_size, self.size)
        else:
            self.size = min(new_size + 2048 * sector_size, self.size)

        if not silent:
            self.out.success("Image size is %dMB" %
                             ((self.size + MB - 1) // MB))

        return part_dev

    def mount(self, mpoint, readonly=False):
        """Mount the image file system under a local directory"""

        assert self.mount_local_support, \
            "MOUNT LOCAL not supported for this build of libguestfs"

        assert self._mount_thread is None, "Image is already mounted"

        self.os.mount(readonly=readonly, silent=True)

        if self.g.mount_local(mpoint, readonly=readonly) == -1:
            return

        # The thread will block in mount_local_run until the file
        self._mount_thread = threading.Thread(target=self.g.mount_local_run)
        self._mount_thread.mpoint = mpoint
        self._mount_thread.start()

    def is_mounted(self):
        """Check if the image is mounted"""

        assert self.mount_local_support, \
            "MOUNT LOCAL not supported for this build of libguestfs"

        if self._mount_thread is None:
            return False

        # Wait for 0.1 second to avoid race conditions if the thread is in an
        # initialization state but not alive yet.
        self._mount_thread.join(0.1)

        return self._mount_thread.is_alive()

    def umount(self, lazy=False):
        """umount the previously mounted image file system"""

        assert self.mount_local_support, \
            "MOUNT LOCAL not supported for this build of libguestfs"

        assert self._mount_thread is not None, "Image is not mounted"

        try:
            # Maybe the image was umounted externally
            if not self._mount_thread.is_alive():
                self._mount_thread = None
                return True

            try:
                args = (['-l'] if lazy else []) + [self._mount_thread.mpoint]
                get_command('umount')(*args)
            except:
                return False

            # Wait for a little while. If the image is umounted,
            # mount_local_run should have terminated
            self._mount_thread.join(5)

            if self._mount_thread.is_alive():
                raise FatalError('Unable to join the mount thread')

            self._mount_thread = None
            return True
        finally:
            self.os.umount()

    def dump(self, outfile):
        """Dumps the content of the image into a file.

        This method will only dump the actual payload, found by reading the
        partition table. Empty space in the end of the device will be ignored.
        """
        MB = 2 ** 20
        blocksize = 2 ** 22  # 4MB
        progr_size = (self.size + MB - 1) // MB  # in MB
        progressbar = self.out.Progress(progr_size, "Dumping image file", 'mb')

        with self.raw_device() as raw:
            with open(raw, 'rb') as src:
                with open(outfile, "wb") as dst:
                    left = self.size
                    offset = 0
                    progressbar.next()
                    while left > 0:
                        length = min(left, blocksize)
                        sent = sendfile(dst.fileno(), src.fileno(), offset,
                                        length)

                        # Workaround for python-sendfile API change. In
                        # python-sendfile 1.2.x (py-sendfile) the returning
                        # value of sendfile is a tuple, where in version 2.x
                        # (pysendfile) it is just a single integer.
                        if isinstance(sent, tuple):
                            sent = sent[1]

                        offset += sent
                        left -= sent
                        progressbar.goto((self.size - left) // MB)

        progressbar.success('image file %s was successfully created' % outfile)

    def md5(self):
        """Computes the MD5 checksum of the image"""

        MB = 2 ** 20
        blocksize = 2 ** 22  # 4MB
        progr_size = ((self.size + MB - 1) // MB)  # in MB
        progressbar = self.out.Progress(progr_size, "Calculating md5sum", 'mb')
        md5 = hashlib.md5()

        with self.raw_device() as raw:
            with open(raw, "rb") as src:
                left = self.size
                while left > 0:
                    length = min(left, blocksize)
                    data = src.read(length)
                    md5.update(data)
                    left -= length
                    progressbar.goto((self.size - left) // MB)

        checksum = md5.hexdigest()
        progressbar.success(checksum)

        return checksum


# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module hosts OS-specific code for OpenBSD."""

import re

from image_creator.os_type.bsd import Bsd


class Openbsd(Bsd):
    """OS class for OpenBSD"""

    def _check_enabled_sshd(self):
        """Check if the ssh daemon is enabled at boot"""

        sshd_enabled = True
        sshd_service = re.compile(r'^sshd_flags=')
        sshd_no = re.compile(r"^sshd_flags=(['\"]?)NO\1$")

        for rc_conf in ('/etc/rc.conf', '/etc/rc.conf.local'):
            if not self.image.g.is_file(rc_conf):
                self.out.warn("File: `%s' does not exist!" % rc_conf)
                continue

            for line in self.image.g.cat(rc_conf).splitlines():
                line = line.split('#')[0].strip()
                if sshd_service.match(line):
                    sshd_enabled = sshd_no.match(line) is None

        return sshd_enabled

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2015 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module is the entrance point for the non-interactive version of the
snf-image-creator program.
"""

from __future__ import unicode_literals

from image_creator import __version__ as version
from image_creator.disk import Disk
from image_creator.util import FatalError
from image_creator.output.cli import SilentOutput, SimpleOutput, \
    OutputWthProgress
from image_creator.output.composite import CompositeOutput
from image_creator.output.syslog import SyslogOutput
from image_creator.kamaki_wrapper import Kamaki, ClientError, CONTAINER
import sys
import os
import argparse
import StringIO
import signal
import json
import textwrap
import tempfile
import subprocess
import time
import re


class CheckWritableDir(argparse.Action):
    """Check if a directory is writable"""
    def __call__(self, parser, namespace, value, option):

        if getattr(namespace, self.dest):
            raise argparse.ArgumentError(self,
                                         "Argument defined multiple times")

        dirname = os.path.dirname(value)
        name = os.path.basename(value)

        if len(dirname) == 0:
            dirname = '.'

        if dirname and not os.path.isdir(dirname):
            raise argparse.ArgumentError(
                self, "`%s' is not an existing directory" % dirname)
        if not os.access(dirname, os.W_OK):
            raise argparse.ArgumentError(
                self, "`%s' is not writable" % dirname)
        if not name:
            raise argparse.ArgumentError(
                self, "`%s' is not a valid file name" % dirname)

        setattr(namespace, self.dest, value.decode(sys.stdin.encoding))


class AddKeyValue(argparse.Action):
    """Add key value options"""
    def __call__(self, parser, namespace, value, option):

        dest = getattr(namespace, self.dest)
        if not dest:
            dest = {}

        value = value.decode(sys.stdin.encoding)
        try:
            key, val = value.split('=', 1)
        except ValueError:
            raise argparse.ArgumentError(
                self, "`%s' is not in KEY=VALUE format." % value)
        dest[key.upper()] = val

        setattr(namespace, self.dest, dest)


def parse_options():
    """Parse input parameters"""
    description = "Command-line tool for creating OS images"
    parser = argparse.ArgumentParser(version=version, description=description)

    parser.add_argument("source", metavar="SOURCE",
                        help="Image file, block device or /")
    parser.add_argument(
        "-a", "--authentication-url", dest="url", default=None,
        help="use this authentication URL when uploading/registering images")

    parser.add_argument(
        "--add-timestamp", dest="timestamp", default=False,
        help="Add a timestamp when outputting messages", action="store_true")

    parser.add_argument(
        "--allow-unsupported", dest="allow_unsupported",
        help="proceed with the image creation even if the media is not "
        "supported", default=False, action="store_true")

    parser.add_argument(
        "-c", "--cloud", dest="cloud", default=None,
        help="use this saved cloud account to authenticate against a cloud "
             "when uploading/registering images")

    parser.add_argument(
        "--container", dest="container", default=CONTAINER,
        help="Upload files to CONTAINER [default: %s]" % CONTAINER)

    parser.add_argument(
        "--disable-sysprep", dest="disabled_syspreps",
        help="prevent SYSPREP operation from running on the input media",
        default=[], action="append", metavar="SYSPREP")

    parser.add_argument(
        "--enable-sysprep", dest="enabled_syspreps", default=[],
        help="run SYSPREP operation on the input media", action="append",
        metavar="SYSPREP")

    parser.add_argument(
        "-f", "--force", dest="force", default=False, action="store_true",
        help="overwrite output files if they exist")

    parser.add_argument(
        "--host-run", dest="host_run", default=[],
        help="mount the media in the host and run a script against the guest "
             "media. This option may be defined multiple times. The script's "
             "working directory will be the guest's root directory. BE "
             "CAREFUL! DO NOT USE ABSOLUTE PATHS INSIDE THE SCRIPT! YOU MAY "
             "HARM YOUR SYSTEM!",
        metavar="SCRIPT", action="append")

    parser.add_argument(
        "--install-virtio", dest="virtio", metavar="DIR",
        help="install VirtIO drivers hosted under DIR (Windows only)")

    parser.add_argument(
        "-m", "--metadata", dest="metadata", default={}, action=AddKeyValue,
        help="add custom KEY=VALUE metadata to the image", metavar="KEY=VALUE")

    parser.add_argument(
        "--no-snapshot", dest="snapshot", default=True,
        help="don't snapshot the input media. (THIS IS DANGEROUS AS IT WILL "
             "ALTER THE ORIGINAL MEDIA!!!)", action="store_false")

    parser.add_argument("--no-sysprep", dest="sysprep", default=True,
                        help="don't perform any system preparation operation",
                        action="store_false")

    parser.add_argument("-o", "--outfile", dest="outfile", default=None,
                        action=CheckWritableDir, metavar="FILE",
                        help="dump image to FILE")

    parser.add_argument("--print-metadata", dest="print_metadata",
                        action='store_true', default=False,
                        help="print the detected image metadata")

    parser.add_argument(
        "--print-syspreps", dest="print_syspreps", default=False,
        help="print the enabled and disabled system preparation operations "
             "for this input media", action="store_true")

    parser.add_argument(
        "--print-sysprep-params", dest="print_sysprep_params", default=False,
        help="print the defined system preparation parameters for this input "
        "media", action="store_true")

    parser.add_argument("--public", dest="public", default=False,
                        help="register image with the cloud as public",
                        action="store_true")

    parser.add_argument("-r", "--register", dest="register", default=False,
                        metavar="IMAGENAME",
                        help="register the image with a cloud as IMAGENAME")

    parser.add_argument("-s", "--silent", dest="silent", default=False,
                        help="output only errors", action="store_true")

    parser.add_argument('--syslog', dest="syslog", default=False,
                        help="log to syslog", action="store_true")

    parser.add_argument("--sysprep-param", dest="sysprep_params", default={},
                        help="add KEY=VALUE system preparation parameter",
                        action=AddKeyValue)

    parser.add_argument(
        "-t", "--token", dest="token", default=None,
        help="use this authentication token when uploading/registering images")

    parser.add_argument("--tmpdir", dest="tmp", default=None, metavar="DIR",
                        help="create large temporary image files under DIR")

    parser.add_argument(
        "-u", "--upload", dest="upload", default=None, metavar="FILENAME",
        help="upload the image to the cloud with name FILENAME")

    options = parser.parse_args()

    if not os.path.exists(options.source):
        parser.error("Input media `%s' is not accessible" % options.source)

    if options.register and not options.upload:
        parser.error("You also need to set -u when -r option is set")

    if options.upload and (options.token is None or options.url is None) and \
            options.cloud is None:

        parser.error("Image uploading cannot be performed. You need to either "
                     "specify an authentication URL and token pair or an "
                     "available cloud name.")

    if options.tmp is not None and not os.path.isdir(options.tmp):
        parser.error("The directory `%s' specified with --tmpdir is not valid"
                     % options.tmp)

    # Convert input attributes to unicode
    for opt in ('url', 'cloud', 'container', 'outfile', 'register', 'token',
                'tmp', 'upload', 'virtio'):
        attr = getattr(options, opt)
        if attr:
            setattr(options, opt, attr.decode(sys.stdin.encoding))

    options.host_run = [h.decode(sys.stdin.encoding) for h in options.host_run]

    metadata_regexp = re.compile('^[A-Za-z_]+$')
    for m in options.metadata:
        if not re.match(metadata_regexp, m):
            parser.error("Metadata key: `%s' is not valid. Allowed characters "
                         "for key: [a-zA-Z0-9_]." % m)

    if options.virtio is not None:
        options.sysprep_params['virtio'] = options.virtio

    if options.outfile is None and not options.upload and not \
            options.print_syspreps and not options.print_sysprep_params \
            and not options.print_metadata:
        parser.error("At least one of `-o', `-u', `--print-syspreps', "
                     "`--print-sysprep-params' or `--print-metadata' must be "
                     "set")

    if not options.force and options.outfile is not None and \
            os.path.realpath(options.outfile) != '/dev/null':
        for extension in ('', '.meta', '.md5sum'):
            filename = "%s%s" % (options.outfile, extension)
            if os.path.exists(filename):
                parser.error("Output file `%s' exists (use --force to "
                             "overwrite it)." % filename)

    return options


def image_creator(options, out):
    """snf-mkimage main function"""

    if os.geteuid() != 0:
        raise FatalError("You must run %s as root"
                         % os.path.basename(sys.argv[0]))

    # Check if the authentication info is valid. The earlier the better
    if options.token is not None and options.url is not None:
        try:
            account = Kamaki.create_account(options.url, options.token)
            if account is None:
                raise FatalError("The authentication token and/or URL you "
                                 "provided is not valid!")
            else:
                kamaki = Kamaki(account, out)
        except ClientError as e:
            raise FatalError("Astakos client: %d %s" % (e.status, e.message))
    elif options.cloud:
        avail_clouds = Kamaki.get_clouds()
        if options.cloud not in avail_clouds.keys():
            raise FatalError(
                "Cloud: `%s' does not exist.\n\nAvailable clouds:\n\n\t%s\n"
                % (options.cloud, "\n\t".join(avail_clouds.keys())))
        try:
            account = Kamaki.get_account(options.cloud)
            if account is None:
                raise FatalError(
                    "Cloud: `%s' exists but is not valid!" % options.cloud)
            else:
                kamaki = Kamaki(account, out)
        except ClientError as e:
            raise FatalError("Astakos client: %d %s" % (e.status, e.message))

    if options.upload and not options.force:
        if kamaki.object_exists(options.container, options.upload):
            raise FatalError("Remote storage service object: `%s' exists "
                             "(use --force to overwrite it)." % options.upload)
        if kamaki.object_exists(options.container,
                                "%s.md5sum" % options.upload):
            raise FatalError("Remote storage service object: `%s.md5sum' "
                             "exists (use --force to overwrite it)." %
                             options.upload)

    if options.register and not options.force:
        if kamaki.object_exists(options.container, "%s.meta" % options.upload):
            raise FatalError("Remote storage service object `%s.meta' exists "
                             "(use --force to overwrite it)." % options.upload)

    disk = Disk(options.source, out, options.tmp)

    def signal_handler(signum, frame):
        disk.cleanup()

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    try:
        # There is no need to snapshot the media if it was created by the Disk
        # instance as a temporary object.
        device = disk.file if not options.snapshot else disk.snapshot()
        image = disk.get_image(device, sysprep_params=options.sysprep_params)

        if image.is_unsupported() and not options.allow_unsupported:
            raise FatalError(
                "The media seems to be unsupported.\n\n" +
                textwrap.fill("To create an image from an unsupported media, "
                              "you'll need to use the`--allow-unsupported' "
                              "command line option. Using this is highly "
                              "discouraged, since the resulting image will "
                              "not be cleared out of sensitive data and will "
                              "not get customized during the deployment."))

        if len(options.host_run) != 0 and not image.mount_local_support:
            raise FatalError("Running scripts against the guest media is not "
                             "supported for this build of libguestfs.")

        if len(options.host_run) != 0:
            for script in options.host_run:
                if not os.path.isfile(script):
                    raise FatalError("File: `%s' does not exist." % script)
                if not os.access(script, os.X_OK):
                    raise FatalError("File: `%s' is not executable." % script)

        for name in options.disabled_syspreps:
            sysprep = image.os.get_sysprep_by_name(name)
            if sysprep is not None:
                image.os.disable_sysprep(sysprep)
            else:
                out.warn("Sysprep: `%s' does not exist. Can't disable it." %
                         name)

        for name in options.enabled_syspreps:
            sysprep = image.os.get_sysprep_by_name(name)
            if sysprep is not None:
                image.os.enable_sysprep(sysprep)
            else:
                out.warn("Sysprep: `%s' does not exist. Can't enable it." %
                         name)
        if image.is_unsupported():
            image.meta['EXCLUDE_ALL_TASKS'] = "yes"

        # Add command line metadata to the collected ones...
        image.meta.update(options.metadata)

        if options.print_syspreps:
            image.os.print_syspreps()
            out.info()

        if options.print_sysprep_params:
            image.os.print_sysprep_params()
            out.info()

        if options.print_metadata:
            image.os.print_metadata()
            out.info()

        if options.outfile is None and not options.upload:
            return 0

        if options.virtio is not None and \
                hasattr(image.os, 'install_virtio_drivers'):
            image.os.install_virtio_drivers()

        if len(options.host_run) != 0:
            # Export image metadata as environment variables to make them
            # visible by the scripts
            for key, value in image.meta.items():
                os.environ["SNF_IMAGE_CREATOR_METADATA_%s" % key] = str(value)

            out.info("Running scripts on the input media:")
            mpoint = tempfile.mkdtemp()
            try:
                image.mount(mpoint)
                if not image.is_mounted():
                    raise FatalError("Mounting the media on the host failed.")
                try:
                    size = len(options.host_run)
                    cnt = 1
                    for script in options.host_run:
                        script = os.path.abspath(script)
                        out.info(("(%d/%d)" % (cnt, size)).ljust(7), False)
                        out.info("Running `%s'" % script)
                        ret = subprocess.Popen([script], cwd=mpoint).wait()
                        if ret != 0:
                            raise FatalError("Script: `%s' failed (rc=%d)" %
                                             (script, ret))
                        cnt += 1
                finally:
                    while not image.umount():
                        out.warn("Unable to umount the media. Retrying ...")
                        time.sleep(1)
                    out.info()
            finally:
                os.rmdir(mpoint)

        if options.sysprep:
            image.os.do_sysprep()

        checksum = image.md5()

        metastring = json.dumps(
            {'properties': image.meta, 'disk-format': 'diskdump'},
            ensure_ascii=False)

        if options.outfile is not None:
            if os.path.realpath(options.outfile) == '/dev/null':
                out.warn('Not dumping file to /dev/null')
            else:
                image.dump(options.outfile)

                out.info('Dumping metadata file ...', False)
                with open('%s.%s' % (options.outfile, 'meta'), 'w') as f:
                    f.write(metastring)
                out.success('done')

                out.info('Dumping md5sum file ...', False)
                with open('%s.%s' % (options.outfile, 'md5sum'), 'w') as f:
                    f.write('%s %s\n' % (checksum,
                                         os.path.basename(options.outfile)))
                out.success('done')

        out.info()
        try:
            if options.upload:
                out.info("Uploading image to the storage service:")
                with image.raw_device() as raw:
                    with open(raw, 'rb') as f:
                        remote = kamaki.upload(
                            f, image.size, options.upload, options.container,
                            "(1/3)  Calculating block hashes",
                            "(2/3)  Uploading missing blocks")

                out.info("(3/3)  Uploading md5sum file ...", False)
                md5sumstr = '%s %s\n' % (checksum,
                                         os.path.basename(options.upload))
                kamaki.upload(StringIO.StringIO(md5sumstr),
                              size=len(md5sumstr),
                              remote_path="%s.%s" % (options.upload, 'md5sum'),
                              container=options.container,
                              content_type="text/plain")
                out.success('done')
                out.info()

            if options.register:
                img_type = 'public' if options.public else 'private'
                out.info('Registering %s image with the compute service ...'
                         % img_type, False)
                result = kamaki.register(options.register, remote,
                                         image.meta, options.public)
                out.success('done')
                out.info("Uploading metadata file ...", False)
                metastring = unicode(json.dumps(result, ensure_ascii=False,
                                                indent=4))
                kamaki.upload(StringIO.StringIO(metastring.encode('utf8')),
                              size=len(metastring),
                              remote_path="%s.%s" % (options.upload, 'meta'),
                              container=options.container,
                              content_type="application/json")
                out.success('done')
                if options.public:
                    out.info("Sharing md5sum file ...", False)
                    kamaki.share("%s.md5sum" % options.upload)
                    out.success('done')
                    out.info("Sharing metadata file ...", False)
                    kamaki.share("%s.meta" % options.upload)
                    out.success('done')
                out.result(json.dumps(result, indent=4, ensure_ascii=False))
                out.info()
        except ClientError as e:
            raise FatalError("Service client: %d %s" % (e.status, e.message))

    finally:
        out.info('cleaning up ...')
        disk.cleanup()

    out.success("snf-image-creator exited without errors")

    return 0


def main():
    """Main entry point"""
    options = parse_options()

    if options.silent:
        out = SilentOutput(colored=sys.stderr.isatty(),
                           timestamp=options.timestamp)
    elif sys.stderr.isatty():
        out = OutputWthProgress(timestamp=options.timestamp)
    else:
        out = SimpleOutput(colored=False, timestamp=options.timestamp)

    if options.syslog:
        out = CompositeOutput([out, SyslogOutput()])

    title = 'snf-image-creator %s' % version
    sys.stderr.write(title + '\n')
    sys.stderr.write(('=' * len(title)) + '\n')

    try:
        sys.exit(image_creator(options, out))
    except FatalError as e:
        out.error(e)
        sys.exit(1)

if __name__ == '__main__':
    main()

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2016 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from os.path import dirname, abspath, join
from setuptools import setup, find_packages
from imp import load_source

CWD = dirname(abspath(__file__))
README = join(CWD, 'README.md')
VERSION = join(CWD, 'image_creator', 'version.py')

setup(
    name='snf_image_creator',
    version=getattr(load_source('version', VERSION), "__version__"),
    description='Command line tool for creating images',
    long_description=open(README).read(),
    url='https://github.com/grnet/snf-image',
    download_url='https://pypi.python.org/pypi/snf_image_creator',
    author='Synnefo development team',
    author_email='synnefo-devel@googlegroups.com',
    maintainer='Synnefo development team',
    maintainer_email='synnefo-devel@googlegroups.com',
    license='GNU GPLv3',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['sh', 'ansicolors', 'progress>=1.0.2', 'kamaki>=0.9',
                      'argparse'],
    # Unresolvable dependencies:
    #   pysendfile|py-sendfile, hivex, guestfs, parted, rsync,
    entry_points={
        'console_scripts': [
                'snf-mkimage = image_creator.main:main',
                'snf-image-creator = image_creator.dialog_main:main']
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: Console :: Curses',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7'],
    keywords='cloud IaaS OS images'
)
# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module hosts OS-specific code for Arch Linux."""

from image_creator.os_type.linux import Linux

import re


class Archlinux(Linux):
    """OS class for Arch Linux"""

    def _do_collect_metadata(self):
        """Collect metadata about the OS"""
        super(Archlinux, self)._do_collect_metadata()

        local_be = '/var/lib/pacman/local'

        if not self.image.g.is_dir(local_be):
            self.out.warn("Directory: `%s' does not exist!" % local_be)
            return

        kernel_regexp = re.compile(r'linux-(lts-)?(\d+[\.\d+]*-\d+)')
        for f in self.image.g.readdir(local_be):
            match = kernel_regexp.match(f['name'])
            if match:
                lts = match.group(1) is not None
                version = match.group(2)
                self.meta['KERNEL'] = "%s%s" % (version, " LTS" if lts else "")

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides an interface for the rsync utility."""

import subprocess
import time
import signal

from image_creator.util import FatalError


class Rsync(object):
    """Wrapper class for the rsync command"""

    def __init__(self, output):
        """Create an instance """
        self._out = output
        self._exclude = []
        self._options = ['-v']

    def archive(self):
        """Enable the archive option"""
        self._options.append('-a')
        return self

    def xattrs(self):
        """Preserve extended attributes"""
        self._options.append('-X')
        return self

    def hard_links(self):
        """Preserve hard links"""
        self._options.append('-H')
        return self

    def acls(self):
        """Preserve ACLs"""
        self._options.append('-A')
        return self

    def sparse(self):
        """Handle sparse files efficiently"""
        self._options.append('-S')
        return self

    def exclude(self, pattern):
        """Add an exclude pattern"""
        self._exclude.append(pattern)
        return self

    def reset(self):
        """Reset all rsync options"""
        self._exclude = []
        self._options = ['-v']

    def run(self, src, dest, slabel='source', dlabel='destination'):
        """Run the actual command"""
        cmd = []
        cmd.append('rsync')
        cmd.extend(self._options)
        for i in self._exclude:
            cmd.extend(['--exclude', i])

        self._out.info("Calculating total number of %s files ..." % slabel,
                       False)

        # If you don't specify a destination, rsync will list the source files.
        dry_run = subprocess.Popen(cmd + [src], shell=False,
                                   stdout=subprocess.PIPE, bufsize=0)
        try:
            total = 0
            for _ in iter(dry_run.stdout.readline, b''):
                total += 1
        finally:
            dry_run.communicate()
            if dry_run.returncode != 0:
                raise FatalError("rsync failed")

        self._out.success("%d" % total)

        progress = self._out.Progress(total, "Copying files to %s" % dlabel)
        run = subprocess.Popen(cmd + [src, dest], shell=False,
                               stdout=subprocess.PIPE, bufsize=0)
        try:
            t = time.time()
            i = 0
            for _ in iter(run.stdout.readline, b''):
                i += 1
                current = time.time()
                if current - t > 0.1:
                    t = current
                    progress.goto(i)

            progress.success('done')

        finally:
            def handler(signum, frame):
                """Signal handler"""
                run.terminate()
                time.sleep(1)
                run.poll()
                if run.returncode is None:
                    run.kill()
                run.wait()

            signal.signal(signal.SIGALRM, handler)
            signal.alarm(2)
            run.communicate()
            signal.alarm(0)
            if run.returncode != 0:
                raise FatalError("rsync failed")

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2016 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module hosts OS-specific code for Linux."""

from image_creator.os_type.unix import Unix, sysprep, add_sysprep_param

import os
import re
import pkg_resources
import tempfile

X2GO_DESKTOPSESSIONS = {
    'CINNAMON': 'cinnamon',
    'KDE': 'startkde',
    'GNOME': 'gnome-session',
    'MATE': 'mate-session',
    'XFCE': 'xfce4-session',
    'LXDE': 'startlxde',
    'TRINITY': 'starttrinity',
    'UNITY': 'unity',
}

X2GO_EXECUTABLE = "x2goruncommand"

DISTRO_ORDER = {
    "ubuntu": 80,
    "linuxmint": 75,
    "debian": 70,
    "rhel": 60,
    "fedora": 58,
    "centos": 55,
    "scientificlinux": 50,
    "sles": 45,
    "opensuse": 44,
    "archlinux": 40,
    "gentoo": 35,
    "slackware": 30,
    "oraclelinux": 28,
    "mageia": 20,
    "mandriva": 19,
    "cirros": 15,
    "pardus": 10
}


class Linux(Unix):
    """OS class for Linux"""
    @add_sysprep_param(
        'bootmenu_timeout', 'posint', 10, "Boot menu timeout in seconds")
    @add_sysprep_param(
        'powerbtn_action', 'string', '/sbin/shutdown -h now',
        "The action that should be executed if the power button is pressed")
    def __init__(self, image, **kwargs):
        super(Linux, self).__init__(image, **kwargs)
        self._uuid = dict()
        self._persistent = re.compile('/dev/[hsv]d[a-z][1-9]*')

    @sysprep('Removing user accounts with id greater that 1000', enabled=False)
    def _remove_user_accounts(self):
        """Remove all user accounts with id greater than 1000"""

        removed_users = {}

        # Remove users from /etc/passwd
        if self.image.g.is_file('/etc/passwd'):
            passwd = []
            metadata_users = self.meta['USERS'].split() \
                if 'USERS' in self.meta else []
            for line in self.image.g.cat('/etc/passwd').splitlines():
                fields = line.split(':')
                if int(fields[2]) > 1000:
                    removed_users[fields[0]] = fields
                    # remove it from the USERS metadata too
                    if fields[0] in metadata_users:
                        metadata_users.remove(fields[0])
                else:
                    passwd.append(':'.join(fields))

            self.meta['USERS'] = " ".join(metadata_users)

            # Delete the USERS metadata if empty
            if not len(self.meta['USERS']):
                del self.meta['USERS']

            self.image.g.write('/etc/passwd', '\n'.join(passwd) + '\n')
        else:
            self.out.warn("File: `/etc/passwd' is missing. "
                          "No users were deleted")
            return

        if self.image.g.is_file('/etc/shadow'):
            # Remove the corresponding /etc/shadow entries
            shadow = []
            for line in self.image.g.cat('/etc/shadow').splitlines():
                fields = line.split(':')
                if fields[0] not in removed_users:
                    shadow.append(':'.join(fields))
            self.image.g.write('/etc/shadow', "\n".join(shadow) + '\n')
        else:
            self.out.warn("File: `/etc/shadow' is missing.")

        if self.image.g.is_file('/etc/group'):
            # Remove the corresponding /etc/group entries
            group = []
            for line in self.image.g.cat('/etc/group').splitlines():
                fields = line.split(':')
                # Remove groups tha have the same name as the removed users
                if fields[0] not in removed_users:
                    group.append(':'.join(fields))
            self.image.g.write('/etc/group', '\n'.join(group) + '\n')

        # Remove home directories
        for home in [field[5] for field in removed_users.values()]:
            if self.image.g.is_dir(home) and home.startswith('/home/'):
                self.image.g.rm_rf(home)

    @sysprep('Cleaning up password & locking all user accounts')
    def _cleanup_passwords(self):
        """Remove all passwords and lock all user accounts"""

        shadow = []

        for line in self.image.g.cat('/etc/shadow').splitlines():
            fields = line.split(':')
            if fields[1] not in ('*', '!'):
                fields[1] = '!'

            shadow.append(":".join(fields))

        self.image.g.write('/etc/shadow', "\n".join(shadow) + '\n')

        # Remove backup file for /etc/shadow
        self.image.g.rm_rf('/etc/shadow-')

    @sysprep('Fixing acpid powerdown action')
    def _fix_acpid(self):
        """Replace acpid powerdown action scripts to immediately shutdown the
        system without checking if a GUI is running.
        """

        powerbtn_action = self.sysprep_params['powerbtn_action'].value

        events_dir = '/etc/acpi/events'
        if not self.image.g.is_dir(events_dir):
            self.out.warn("No acpid event directory found")
            return

        event_exp = re.compile('event=(.+)', re.I)
        action_exp = re.compile('action=(.+)', re.I)
        for events_file in self.image.g.readdir(events_dir):
            if events_file['ftyp'] != 'r':
                continue

            event = -1
            action = -1
            fullpath = "%s/%s" % (events_dir, events_file['name'])
            content = self.image.g.cat(fullpath).splitlines()
            for i in xrange(len(content)):
                if event_exp.match(content[i]):
                    event = i
                elif action_exp.match(content[i]):
                    action = i

            if event == -1:
                continue

            if action == -1:
                self.out.warn("Corrupted acpid event file: `%s'" % fullpath)
                continue

            entry = content[event].split('=')[1].strip()
            if entry in ("button[ /]power", "button/power.*"):
                content[action] = "action=%s" % powerbtn_action
                self.image.g.write(fullpath, "\n".join(content) +
                                   '\n\n### Edited by snf-image-creator ###\n')
                return
            elif entry == ".*":
                self.out.warn("Found action `.*'. Don't know how to handle "
                              "this. Please edit `%s' image file manually to "
                              "make the system immediately shutdown when an "
                              "power button ACPI event occurs." %
                              content[action].split('=')[1].strip())
                return

        self.out.warn("No acpi power button event found!")

    @sysprep('Removing persistent network interface names')
    def _remove_persistent_net_rules(self):
        """Remove udev rules that will keep network interface names persistent
        after hardware changes and reboots. Those rules will be created again
        the next time the image runs.
        """

        rule_file = '/etc/udev/rules.d/70-persistent-net.rules'
        if self.image.g.is_file(rule_file):
            self.image.g.rm(rule_file)

    @sysprep('Removing swap entry from fstab')
    def _remove_swap_entry(self):
        """Remove swap entry from /etc/fstab. If swap is the last partition
        then the partition will be removed when shrinking is performed. If the
        swap partition is not the last partition in the disk or if you are not
        going to shrink the image you should probably disable this.
        """

        if not self.image.g.is_file('/etc/fstab'):
            self.out.warn("File: `/etc/fstab' is missing. No entry removed!")
            return

        new_fstab = ""
        fstab = self.image.g.cat('/etc/fstab')
        for line in fstab.splitlines():

            entry = line.split('#')[0].strip().split()
            if len(entry) == 6 and entry[2] == 'swap':
                continue

            new_fstab += "%s\n" % line

        self.image.g.write('/etc/fstab', new_fstab)

    @sysprep('Change boot menu timeout to %(bootmenu_timeout)s seconds')
    def _change_bootmenu_timeout(self):
        """Change the boot menu timeout to the one specified by the namesake
        system preparation parameter.
        """

        timeout = self.sysprep_params['bootmenu_timeout'].value

        if self.image.g.is_file('/etc/default/grub'):
            self.image.g.aug_init('/', 0)
            try:
                self.image.g.aug_set('/files/etc/default/grub/GRUB_TIMEOUT',
                                     str(timeout))
            finally:
                self.image.g.aug_save()
                self.image.g.aug_close()

        def replace_timeout(remote, regexp, timeout):
            """Replace the timeout value from a config file"""
            tmpfd, tmp = tempfile.mkstemp()
            try:
                for line in self.image.g.cat(remote).splitlines():
                    if regexp.match(line):
                        line = re.sub(r'\d+', str(timeout), line)
                    os.write(tmpfd, line + '\n')
                os.close(tmpfd)
                tmpfd = None
                self.image.g.upload(tmp, remote)
            finally:
                if tmpfd is not None:
                    os.close(tmpfd)
                os.unlink(tmp)

        grub1_config = '/boot/grub/menu.lst'
        grub2_config = '/boot/grub/grub.cfg'
        syslinux_config = '/boot/syslinux/syslinux.cfg'

        if self.image.g.is_file(grub1_config):
            regexp = re.compile(r'^\s*timeout\s+\d+\s*$')
            replace_timeout(grub1_config, regexp, timeout)
        elif self.image.g.is_file(grub2_config):
            regexp = re.compile(r'^\s*set\s+timeout=\d+\s*$')
            replace_timeout(grub2_config, regexp, timeout)

        if self.image.g.is_file(syslinux_config):
            regexp = re.compile(r'^\s*TIMEOUT\s+\d+\s*$', re.IGNORECASE)
            # In syslinux the timeout unit is 0.1 seconds
            replace_timeout(syslinux_config, regexp, timeout * 10)

    @sysprep('Replacing fstab & grub non-persistent device references')
    def _use_persistent_block_device_names(self):
        """Scan fstab & grub configuration files and replace all non-persistent
        device references with UUIDs.
        """

        if not self.image.g.is_file('/etc/fstab'):
            self.out.warn("Omitted! File: `/etc/fstab' does not exist")

        # convert all devices in fstab to persistent
        persistent_root = self._persistent_fstab()

        # convert root device in grub1 to persistent
        self._persistent_grub1(persistent_root)

        # convert root device in syslinux to persistent
        self._persistent_syslinux(persistent_root)

    @sysprep('Disabling IPv6 privacy extensions',
             display='Disable IPv6 privacy enxtensions')
    def _disable_ipv6_privacy_extensions(self):
        """Disable IPv6 privacy extensions."""

        file_path = '/files/etc/sysctl.conf/net.ipv6.conf.%s.use_tempaddr'
        dir_path = '/files/etc/sysctl.d/*/net.ipv6.conf.%s.use_tempaddr'

        self.image.g.aug_init('/', 0)
        try:
            default = self.image.g.aug_match(file_path % 'default') + \
                self.image.g.aug_match(dir_path % 'default')

            all = self.image.g.aug_match(file_path % 'all') + \
                self.image.g.aug_match(dir_path % 'all')

            if len(default) == 0:
                self.image.g.aug_set(file_path % 'default', '0')
            else:
                for token in default:
                    self.image.g.aug_set(token, '0')

            if len(all) == 0:
                self.image.g.aug_set(file_path % 'all', '0')
            else:
                for token in all:
                    self.image.g.aug_set(token, '0')

        finally:
            self.image.g.aug_save()
            self.image.g.aug_close()

    @sysprep('Removing local machine ID configuration file',
             display="Remove local machine ID configuration file")
    def _remove_local_machine_id_configuration_file(self):
        """Remove the /etc/machine-id file if present. This file is used by
        systemd to uniquelly identify systems and will be created automatically
        on the next boot if not present."""

        if self.image.g.is_file('/etc/machine-id'):
            self.image.g.rm('/etc/machine-id')

    def _persistent_grub1(self, new_root):
        """Replaces non-persistent device name occurrences with persistent
        ones in GRUB1 configuration files.
        """
        if self.image.g.is_file('/boot/grub/menu.lst'):
            grub1 = '/boot/grub/menu.lst'
        elif self.image.g.is_file('/etc/grub.conf'):
            grub1 = '/etc/grub.conf'
        else:
            return

        self.image.g.aug_init('/', 0)
        try:
            roots = self.image.g.aug_match(
                '/files%s/title[*]/kernel/root' % grub1)
            for root in roots:
                dev = self.image.g.aug_get(root)
                if not self._is_persistent(dev):
                    # This is not always correct. Grub may contain root entries
                    # for other systems, but we only support 1 OS per hard
                    # disk, so this shouldn't harm.
                    self.image.g.aug_set(root, new_root)
        finally:
            self.image.g.aug_save()
            self.image.g.aug_close()

    def _persistent_syslinux(self, new_root):
        """Replace non-persistent root device name occurrences with persistent
        ones in the syslinux configuration files.
        """

        config = '/boot/syslinux/syslinux.cfg'
        append_regexp = re.compile(
            r'\s*APPEND\s+.*\broot=/dev/[hsv]d[a-z][1-9]*\b', re.IGNORECASE)

        if not self.image.g.is_file(config):
            return

        # There is no augeas lense for syslinux :-(
        tmpfd, tmp = tempfile.mkstemp()
        try:
            for line in self.image.g.cat(config).splitlines():
                if append_regexp.match(line):
                    line = re.sub(r'\broot=/dev/[hsv]d[a-z][1-9]*\b',
                                  'root=%s' % new_root, line)
                os.write(tmpfd, line + '\n')
            os.close(tmpfd)
            tmpfd = None
            self.image.g.upload(tmp, config)
        finally:
            if tmpfd is not None:
                os.close(tmpfd)
            os.unlink(tmp)

    def _persistent_fstab(self):
        """Replaces non-persistent device name occurrences in /etc/fstab with
        persistent ones.
        """
        mpoints = self.image.g.mountpoints()
        if len(mpoints) == 0:
            pass  # TODO: error handling

        device_dict = dict([[mpoint, dev] for dev, mpoint in mpoints])

        root_dev = None
        new_fstab = ""
        fstab = self.image.g.cat('/etc/fstab')
        for line in fstab.splitlines():

            line, dev, mpoint = self._convert_fstab_line(line, device_dict)
            new_fstab += "%s\n" % line

            if mpoint == '/':
                root_dev = dev

        self.image.g.write('/etc/fstab', new_fstab)
        if root_dev is None:
            pass  # TODO: error handling

        return root_dev

    def _convert_fstab_line(self, line, devices):
        """Replace non-persistent device names in an fstab line to their UUID
        equivalent
        """
        orig = line
        line = line.split('#')[0].strip()
        if len(line) == 0:
            return orig, "", ""

        entry = line.split()
        if len(entry) != 6:
            self.out.warn("Detected abnormal entry in fstab")
            return orig, "", ""

        dev = entry[0]
        mpoint = entry[1]

        if not self._is_persistent(dev):
            if mpoint in devices:
                dev = "UUID=%s" % self._get_uuid(devices[mpoint])
                entry[0] = dev
            else:
                # comment out the entry
                entry[0] = "#%s" % dev
            return " ".join(entry), dev, mpoint

        return orig, dev, mpoint

    def _do_inspect(self):
        """Run various diagnostics to check if media is supported"""

        self.out.info(
            'Checking if the media contains logical volumes (LVM)...', False)

        has_lvm = True if len(self.image.g.lvs()) else False

        if has_lvm:
            self.out.info()
            self.image.set_unsupported('The media contains logical volumes')
        else:
            self.out.success('no')

    def _do_collect_metadata(self):
        """Collect metadata about the OS"""
        super(Linux, self)._do_collect_metadata()
        users = self._get_passworded_users()
        self.meta["USERS"] = " ".join(users)

        # Delete the USERS metadata if empty
        if not len(self.meta['USERS']):
            self.out.warn("No passworded users found!")
            del self.meta['USERS']

        kernels = []
        for f in self.image.g.ls('/boot'):
            if f.startswith('config-'):
                kernels.append(f[7:])

        if len(kernels):
            kernels.sort(key=pkg_resources.parse_version)
            self.meta['KERNEL'] = kernels[-1]

        distro = self.image.g.inspect_get_distro(self.root)
        major = self.image.g.inspect_get_major_version(self.root)
        if major > 99:
            major = 99
        minor = self.image.g.inspect_get_minor_version(self.root)
        if minor > 99:
            minor = 99
        try:
            self.meta['SORTORDER'] += \
                10000 * DISTRO_ORDER[distro] + 100 * major + minor
        except KeyError:
            pass

        if self.is_enabled('sshd'):
            ssh = []
            opts = self.ssh_connection_options(users)
            for user in opts['users']:
                ssh.append("ssh:port=%d,user=%s" % (opts['port'], user))

            if 'REMOTE_CONNECTION' not in self.meta:
                self.meta['REMOTE_CONNECTION'] = ""
            else:
                self.meta['REMOTE_CONNECTION'] += " "

            if len(ssh):
                self.meta['REMOTE_CONNECTION'] += " ".join(ssh)
            else:
                self.meta['REMOTE_CONNECTION'] += "ssh:port=%d" % opts['port']

            # Check if x2go is installed
            x2go_installed = False
            desktops = set()
            for path in ('/bin', '/usr/bin', '/usr/local/bin'):
                if self.image.g.is_file("%s/%s" % (path, X2GO_EXECUTABLE)):
                    x2go_installed = True
                for name, exe in X2GO_DESKTOPSESSIONS.items():
                    if self.image.g.is_file("%s/%s" % (path, exe)):
                        desktops.add(name)

            if x2go_installed:
                self.meta['REMOTE_CONNECTION'] += " "
                if len(desktops) == 0:
                    self.meta['REMOTE_CONNECTION'] += "x2go"
                else:
                    self.meta['REMOTE_CONNECTION'] += \
                        " ".join(["x2go:session=%s" % d for d in desktops])
        else:
            self.out.warn("OpenSSH Daemon is not configured to run on boot")

    def is_enabled(self, service):
        """Check if a service is enabled to run on boot"""

        systemd_services = '/etc/systemd/system/multi-user.target.wants'
        exec_start = re.compile(r'^\s*ExecStart=.+bin/%s\s?' % service)
        if self.image.g.is_dir(systemd_services):
            for entry in self.image.g.readdir(systemd_services):
                if entry['ftyp'] not in ('l', 'f'):
                    continue
                service_file = "%s/%s" % (systemd_services, entry['name'])
                for line in self.image.g.cat(service_file).splitlines():
                    if exec_start.search(line):
                        return True

        found = set()

        def check_file(path):
            regexp = re.compile(r"[/=\s'\"]%s('\")?\s" % service)
            for line in self.image.g.cat(path).splitlines():
                line = line.split('#', 1)[0].strip()
                if len(line) == 0:
                    continue
                if regexp.search(line):
                    found.add(path)
                    return

        # Check upstart config files under /etc/init
        # Only examine *.conf files
        if self.image.g.is_dir('/etc/init'):
            self._foreach_file('/etc/init', check_file, maxdepth=1,
                               include=r'.+\.conf$')
            if len(found):
                return True

        # Check scripts under /etc/rc[1-5].d/ and /etc/rc.d/rc[1-5].d/
        for conf in ["/etc/%src%d.d" % (d, i) for i in xrange(1, 6)
                     for d in ('', 'rc.d/')]:
            try:
                for entry in self.image.g.readdir(conf):
                    if entry['ftyp'] not in ('l', 'f'):
                        continue
                    check_file("%s/%s" % (conf, entry['name']))

                    if len(found):
                        return True

            except RuntimeError:
                continue

        return False

    def _get_passworded_users(self):
        """Returns a list of non-locked user accounts"""

        if not self.image.g.is_file('/etc/shadow'):
            self.out.warn(
                "Unable to collect user info. File: `/etc/shadow' is missing!")
            return []

        users = []
        regexp = re.compile(r'(\S+):((?:!\S+)|(?:[^!*]\S+)|):(?:\S*:){6}')

        for line in self.image.g.cat('/etc/shadow').splitlines():
            match = regexp.match(line)
            if not match:
                continue

            user, passwd = match.groups()
            if len(passwd) > 0 and passwd[0] == '!':
                self.out.warn("Ignoring locked %s account." % user)
            else:
                users.append(user)

        return users

    def _is_persistent(self, dev):
        """Checks if a device name is persistent."""
        return not self._persistent.match(dev)

    def _get_uuid(self, dev):
        """Returns the UUID corresponding to a device"""
        if dev in self._uuid:
            return self._uuid[dev]

        uuid = self.image.g.vfs_uuid(dev)
        assert len(uuid)
        self._uuid[dev] = uuid
        return uuid

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :
